/*++

Copyright (c) 2003	Microsoft Corporation All Rights Reserved

Module Name:

	BUSENUM.C

Abstract:

	This module contains routines to handle the function driver
	aspect of the bus driver. This sample is functionally
	equivalent to the WDM mxe bus driver.

Environment:

	kernel mode only

--*/

#include "precomp.h"
#include <initguid.h>
#include <wdmguid.h>

#if defined (EVENT_TRACING)
#ifdef offsetof
#undef offsetof
#endif
#include "drv.tmh"
#endif 


#define DRV_VERSION	"1.0"
#define DRV_RELDATE	"02/01/2008"

GLOBALS g = {0};
uint32_t g_mlx4_dbg_flags = 0xffff;
uint32_t g_mlx4_dbg_level = TRACE_LEVEL_INFORMATION;
WCHAR g_wlog_buf[ MAX_LOG_BUF_LEN ];
UCHAR g_slog_buf[ MAX_LOG_BUF_LEN ];
u32 g_time_increment = 0;


#ifndef USE_WDM_INTERRUPTS

typedef struct {
	int					int_num;
	PFDO_DEVICE_DATA	p_fdo;
	struct mlx4_eq *	eq;
} INTERRUPT_DATA, *PINTERRUPT_DATA;

WDF_DECLARE_CONTEXT_TYPE(INTERRUPT_DATA);

NTSTATUS	
EvtEnableInterrupt(	
	IN WDFINTERRUPT  Interrupt,
	IN WDFDEVICE  AssociatedDevice
	)
{
	UNUSED_PARAM(Interrupt);
	UNUSED_PARAM(AssociatedDevice);
	MLX4_ENTER(MLX4_DBG_DRV);
	MLX4_EXIT( MLX4_DBG_DRV );
	return STATUS_SUCCESS;
}

NTSTATUS
EvtDisableInterrupt (
	IN WDFINTERRUPT  Interrupt,
	IN WDFDEVICE  AssociatedDevice
	)
{
	UNUSED_PARAM(Interrupt);
	UNUSED_PARAM(AssociatedDevice);
	MLX4_ENTER(MLX4_DBG_DRV);
	MLX4_EXIT( MLX4_DBG_DRV );
	return STATUS_SUCCESS;
}

BOOLEAN
EvtInterruptIsr(
	IN WDFINTERRUPT  Interrupt,
	IN ULONG  MessageID
	)
{
	BOOLEAN isr_handled = FALSE;
	PINTERRUPT_DATA p_isr_ctx = WdfObjectGetTypedContext( Interrupt, INTERRUPT_DATA );

	UNUSED_PARAM(MessageID);

//	MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV, ("Fdo %p\n", p_isr_ctx->p_fdo));
	if (p_isr_ctx->eq && p_isr_ctx->eq->isr)
		isr_handled = p_isr_ctx->eq->isr( p_isr_ctx->eq->eq_ix, p_isr_ctx->eq->ctx );
	
	return isr_handled;
}

#endif

// In units of ms
uint64_t get_tickcount_in_ms(void)
{
	LARGE_INTEGER	Ticks;

	KeQueryTickCount(&Ticks);
	return Ticks.QuadPart * g_time_increment / 10000; // 10,000 moves from 100ns to ms
}


NTSTATUS
__create_child(
	__in WDFDEVICE	Device,
	__in PWCHAR 	HardwareIds,
	__in PWCHAR 	CompatibleIds,
	__in PWCHAR 	DeviceDescription,
	__in ULONG		SerialNo
	)

/*++

Routine Description:

	The user application has told us that a new device on the bus has arrived.

	We therefore need to create a new PDO, initialize it, add it to the list
	of PDOs for this FDO bus, and then tell Plug and Play that all of this
	happened so that it will start sending prodding IRPs.

--*/

{
	NTSTATUS			status = STATUS_SUCCESS;
	BOOLEAN				unique = TRUE;
	WDFDEVICE			hChild;
	PPDO_DEVICE_DATA	p_pdo;
	PFDO_DEVICE_DATA	p_fdo;

	PAGED_CODE ();
	MLX4_ENTER(MLX4_DBG_DRV);

	//
	// First make sure that we don't already have another device with the
	// same serial number.
	// Framework creates a collection of all the child devices we have
	// created so far. So acquire the handle to the collection and lock
	// it before walking the item.
	//
	p_fdo = FdoGetData(Device);
	hChild = NULL;

	//
	// We need an additional lock to synchronize addition because
	// WdfFdoLockStaticChildListForIteration locks against anyone immediately
	// updating the static child list (the changes are put on a queue until the
	// list has been unlocked).  This type of lock does not enforce our concept
	// of unique IDs on the bus (ie SerialNo).
	//
	// Without our additional lock, 2 threads could execute this function, both
	// find that the requested SerialNo is not in the list and attempt to add
	// it.	If that were to occur, 2 PDOs would have the same unique SerialNo,
	// which is incorrect.
	//
	// We must use a passive level lock because you can only call WdfDeviceCreate
	// at PASSIVE_LEVEL.
	//
	WdfWaitLockAcquire(p_fdo->ChildLock, NULL);
	WdfFdoLockStaticChildListForIteration(Device);

	while ((hChild = WdfFdoRetrieveNextStaticChild(Device,
		hChild, WdfRetrieveAddedChildren)) != NULL) {
		//
		// WdfFdoRetrieveNextStaticChild returns reported and to be reported
		// children (ie children who have been added but not yet reported to PNP).
		//
		// A surprise removed child will not be returned in this list.
		//
		p_pdo = PdoGetData(hChild);
		p_pdo->PdoDevice = hChild;
		p_pdo->p_fdo = p_fdo;

		//
		// It's okay to plug in another device with the same serial number
		// as long as the previous one is in a surprise-removed state. The
		// previous one would be in that state after the device has been
		// physically removed, if somebody has an handle open to it.
		//
		if (SerialNo == p_pdo->SerialNo) {
		unique = FALSE;
		status = STATUS_INVALID_PARAMETER;
		break;
		}
	}

	if (unique) {
		//
		// Create a new child device.  It is OK to create and add a child while
		// the list locked for enumeration.  The enumeration lock applies only
		// to enumeration, not addition or removal.
		//
		status = create_pdo(Device, HardwareIds, CompatibleIds, DeviceDescription, SerialNo, p_fdo->pci_dev.location);
	}

	WdfFdoUnlockStaticChildListFromIteration(Device);
	WdfWaitLockRelease(p_fdo->ChildLock);

	MLX4_EXIT( MLX4_DBG_DRV );
	return status;
}


static 
NTSTATUS
__do_static_enumeration(
	IN WDFDEVICE Device
	)
/*++
Routine Description:

	The routine enables you to statically enumerate child devices
	during start instead of running the enum.exe/notify.exe to
	enumerate mxe devices.

	In order to statically enumerate, user must specify the number
	of mxes in the Mxe Bus driver's device registry. The
	default value is 2.

	You can also configure this value in the Mxe Bus Inf file.

--*/

{
	NTSTATUS status = STATUS_SUCCESS;
	int i;
	int number_of_ib_ports;
	PFDO_DEVICE_DATA p_fdo	= FdoGetData(Device);
	struct mlx4_dev *mdev	= p_fdo->pci_dev.dev;
	BOOLEAN ib_created = FALSE;
	BOOLEAN eth_created = FALSE;

#define MAX_HW_ID_STR 100    
    WCHAR eth_hw_ids[MAX_HW_ID_STR] = {0};

	MLX4_ENTER(MLX4_DBG_DRV);
    
#ifdef HP_PROD
    RtlStringCchPrintfW(eth_hw_ids,MAX_HW_ID_STR,L"%s&%x%x",ETH_HARDWARE_IDS,
                                                            mdev->pdev->sub_system_id,
                                                            mdev->pdev->sub_vendor_id
                                                            );
#else
    RtlStringCchCopyW(eth_hw_ids, MAX_HW_ID_STR, ETH_HARDWARE_IDS);
#endif



	if ( p_fdo->children_created )
		goto end;
	
	// eventually we'll have all information about children in Registry
	// DriverEntry will read it into a Global storage and
	// this routine will create all the children on base on this info
	number_of_ib_ports = mlx4_count_ib_ports(mdev);
	ASSERT(number_of_ib_ports >=0 && number_of_ib_ports <=2);

#if 0
	//For now we it's either IB or ETH, and we always create LLE if it's ETH
	if((number_of_ib_ports > 0) && (mdev->caps.port_type[1] == MLX4_PORT_TYPE_IB) ) {
		status = __create_child(Device, BUS_HARDWARE_IDS, BUS_HARDWARE_DESCRIPTION, 0 );
		if (!NT_SUCCESS(status)) {
			 MLX4_PRINT_EV(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, ("__create_child (ib)failed with 0x%x\n", status));
		}
	}
#endif

	for (i = 1; i <= mdev->caps.num_ports; i++) {
		if (mlx4_is_enabled_port(mdev, i)) {
			if(mlx4_is_eth_port(mdev, i)) {
				status = __create_child(Device, eth_hw_ids, ETH_COMPATIBLE_IDS, ETH_HARDWARE_DESCRIPTION, i);
				if (!NT_SUCCESS(status)) {
					MLX4_PRINT_EV(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, ("%s: __create_child (eth) failed with 0x%x\n", 
						mdev->pdev->name, status));
					break;
				}
				eth_created = TRUE;

				if(mdev->caps.enable_fcox[i])
				{
					status = __create_child(Device, FCOE_HARDWARE_IDS, FCOE_HARDWARE_IDS, FCOE_HARDWARE_DESCRIPTION, mdev->caps.num_ports + i);
					if (!NT_SUCCESS(status)) {
						 MLX4_PRINT_EV(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, ("__create_child (fcoe) failed with 0x%x\n", status));
						 break;
					}
				}
			} else {
				if (eth_created){
					//
					// Illegal configuration the IB should be the first port
					//
					status = STATUS_INVALID_PARAMETER;
					MLX4_PRINT_EV(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, ("%s: __create_child (IB) failed. Invalid configuration, IB should be the first port.", mdev->pdev->name ));
					break;
				}

				if (ib_created)
					continue;

				status = __create_child(Device, BUS_HARDWARE_IDS, BUS_HARDWARE_IDS, BUS_HARDWARE_DESCRIPTION, 0 );
				if (!NT_SUCCESS(status)) {
					MLX4_PRINT_EV(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, ("%s: __create_child (ib)failed with 0x%x\n", mdev->pdev->name, status));
					break;
				}
				ib_created = TRUE;
				p_fdo->pci_dev.is_reset_prohibited = TRUE;
			}
		}
	}

	p_fdo->children_created = TRUE;
	
end:
	MLX4_EXIT( MLX4_DBG_DRV );
	return status;
}

BOOLEAN __read_setup_params(struct pci_dev *pdev, PUNICODE_STRING puvalue)
{    
	NTSTATUS	status;    
	WDFKEY hParamsKey = NULL;
	DECLARE_CONST_UNICODE_STRING(KeyName, L"\\Registry\\Machine\\SOFTWARE\\Mellanox");
	DECLARE_CONST_UNICODE_STRING(PortType, L"PortType");

    UNUSED_PARAM(pdev);

    //
    //  Try to read PortType registry key: (eth,eth),(ib,eth),(ib,ib) from SOFTWARE\MELLANOX
    //  In sucess the read value will replace the current configuraion (coming from INF)
    //  In any case of read failure, just continue with the current PortType configuration. 
    //
	status = WdfRegistryOpenKey(NULL, &KeyName, STANDARD_RIGHTS_ALL, WDF_NO_OBJECT_ATTRIBUTES, &hParamsKey);
	if( !NT_SUCCESS( status ) ) 
	{
		return FALSE;
	}

	status = WdfRegistryQueryUnicodeString(hParamsKey, &PortType, NULL, puvalue);	 
	if( !NT_SUCCESS( status ) ) 
	{
		WdfRegistryClose(hParamsKey);
		return FALSE;
	}

	WdfRegistryClose(hParamsKey);	 
	return TRUE;	
}

NTSTATUS
__read_dev_params(IN WDFDEVICE	Device, struct mlx4_dev_params *dev_params)
{
	NTSTATUS status = STATUS_SUCCESS;
	WDFKEY hKey = NULL;
	WDFKEY hParamsKey = NULL;
	BOOLEAN bRet = FALSE;
	PFDO_DEVICE_DATA p_fdo	= FdoGetData(Device);
	struct pci_dev *pdev = &p_fdo->pci_dev;
	DECLARE_CONST_UNICODE_STRING(Parameters, L"Parameters");
	DECLARE_CONST_UNICODE_STRING(PortType, L"PortType");
	DECLARE_CONST_UNICODE_STRING(EnableFCoX, L"EnableFCoX");

#define  MAX_UVALUE 100
	WCHAR uvalue_data[MAX_UVALUE]={0};
	UNICODE_STRING uvalue;	  
	uvalue.Buffer = uvalue_data;
	uvalue.MaximumLength = MAX_UVALUE;
	uvalue.Length = 0;

	// default values
	dev_params->mod_port_type[0] = MLX4_PORT_TYPE_IB;
	dev_params->mod_port_type[1] = MLX4_PORT_TYPE_IB;

	status = WdfDeviceOpenRegistryKey(Device, PLUGPLAY_REGKEY_DRIVER, 
		STANDARD_RIGHTS_ALL, WDF_NO_OBJECT_ATTRIBUTES, &hKey);
	if( !NT_SUCCESS( status ) ) {
		MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,
			("WdfDeviceOpenRegistryKey(\\Registry\\Machine\\Control\\Class\\...) Failed status = 0x%x\n", status));
		if(status == STATUS_OBJECT_NAME_NOT_FOUND ) {
		
			// This is the case of iSCSI install, we need to continue going up with ethernet
			dev_params->mod_port_type[0] = MLX4_PORT_TYPE_ETH;
			dev_params->mod_port_type[1] = MLX4_PORT_TYPE_ETH;
			status = STATUS_SUCCESS;
			goto err;
		}
		WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_WARN_REG_OPEN_DEV_KEY, 0, 0, 1,
			L"%#x", status );
		goto err;
	}

	status = WdfRegistryOpenKey(hKey, &Parameters, STANDARD_RIGHTS_ALL, WDF_NO_OBJECT_ATTRIBUTES, &hParamsKey);
	if( !NT_SUCCESS( status ) ) {
		MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,("WdfRegistryOpenKey(Prameters) Failed status = 0x%x\n", status));
		WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_WARN_REG_ACTION, 0, 0, 3,
			L"%s", L"WdfRegistryOpenKey", L"%s", Parameters.Buffer, L"%#x", status );
		goto err;
	}

	bRet = __read_setup_params(pdev, &uvalue);
	if (bRet == TRUE)
	{
		status = WdfRegistryAssignValue(hParamsKey, &PortType, REG_SZ,uvalue.Length,uvalue.Buffer); 	   
		if( !NT_SUCCESS( status ) ) {
			MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,("WdfRegistryAssignValue(PortType) Failed status = 0x%x\n", status));
			WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_WARN_REG_ACTION, 0, 0, 3,
				L"%s", L"WdfRegistryAssignValue", L"%s", PortType.Buffer, L"%#x", status );
			goto err;
		}
		uvalue.Length = 0;
	}

	memset(&uvalue_data,0,sizeof(uvalue_data));

	status = WdfRegistryQueryUnicodeString(hParamsKey, &PortType, NULL, &uvalue);
	if (NT_SUCCESS (status)) {
		if (!wcscmp(uvalue_data, L"ib,ib")) {
			dev_params->mod_port_type[0] = MLX4_PORT_TYPE_IB;
			dev_params->mod_port_type[1] = MLX4_PORT_TYPE_IB;
		} else
		if (!wcscmp(uvalue_data, L"ib,eth")) {
			dev_params->mod_port_type[0] = MLX4_PORT_TYPE_IB;
			dev_params->mod_port_type[1] = MLX4_PORT_TYPE_ETH;
		} else
		if (!wcscmp(uvalue_data, L"eth,ib")) {
			dev_params->mod_port_type[0] = MLX4_PORT_TYPE_ETH;
			dev_params->mod_port_type[1] = MLX4_PORT_TYPE_IB;
		} else
		if (!wcscmp(uvalue_data, L"eth,eth")) {
			dev_params->mod_port_type[0] = MLX4_PORT_TYPE_ETH;
			dev_params->mod_port_type[1] = MLX4_PORT_TYPE_ETH;
		} else {
			MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,("%s: Invalid value, PortType = %S\n", pdev->name, uvalue_data));			
			WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_WARN_INVALID_PORT_TYPE_VALUE, 0, 0, 1,
				L"%s",uvalue_data);			
		}
	}
	else {
		MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,
			("WdfRegistryQueryUnicodeString(PortType) Failed status = 0x%x\n", status));
	}

	memset(&uvalue_data,0,sizeof(uvalue_data));

	status = WdfRegistryQueryUnicodeString(hParamsKey, &EnableFCoX, NULL, &uvalue);
	if (NT_SUCCESS (status)) {
		if (!wcscmp(uvalue_data, L"0,0")) {
			dev_params->enable_fcox[0] = 0;
			dev_params->enable_fcox[1] = 0;
		} else
		if (!wcscmp(uvalue_data, L"0,1")) {
			dev_params->enable_fcox[0] = 0;
			dev_params->enable_fcox[1] = 1;
		} else
		if (!wcscmp(uvalue_data, L"1,0")) {
			dev_params->enable_fcox[0] = 1;
			dev_params->enable_fcox[1] = 0;
		} else
		if (!wcscmp(uvalue_data, L"1,1")) {
			dev_params->enable_fcox[0] = 1;
			dev_params->enable_fcox[1] = 1;
		} else {
			MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,("%s: Invalid value, EnableFCoX = %S\n", pdev->name, uvalue_data));			
			WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_WARN_INVALID_PORT_TYPE_VALUE, 0, 0, 1,
				L"%s",uvalue_data);			
		}
	}
	else
	{// default values
		dev_params->enable_fcox[0] = 0;
		dev_params->enable_fcox[1] = 0;
		status = STATUS_SUCCESS;
	}
err:
	if (hKey != NULL) 
		WdfRegistryClose(hKey);

	if (hParamsKey != NULL) 
		WdfRegistryClose(hParamsKey);

	return status;
}

static 
void InitInterface(PFDO_DEVICE_DATA p_fdo)
{
	//
	// prepare MLX4 IB interface
	//

	// fill the header
	p_fdo->bus_ib_ifc.i.Size = sizeof(MLX4_BUS_IB_INTERFACE);
	p_fdo->bus_ib_ifc.i.Version = MLX4_BUS_IB_INTERFACE_VERSION;
	// Let the framework handle reference counting.
	p_fdo->bus_ib_ifc.i.InterfaceReference = WdfDeviceInterfaceReferenceNoOp;
	p_fdo->bus_ib_ifc.i.InterfaceDereference = WdfDeviceInterfaceDereferenceNoOp;

	p_fdo->bus_ib_ifc.pdev = &p_fdo->pci_dev;
	p_fdo->bus_ib_ifc.p_ibdev = p_fdo->pci_dev.ib_dev;
	p_fdo->bus_ib_ifc.pmlx4_dev = to_mdev(p_fdo->pci_dev.ib_dev)->dev;
	if ( mlx4_is_msi(p_fdo->bus_ib_ifc.pmlx4_dev) )
		p_fdo->bus_ib_ifc.n_msi_vectors = 
			mlx4_priv( p_fdo->pci_dev.dev )->eq_table.num_eth_eqs;

	//
	// prepare MLX4 FC interface
	//
	p_fdo->bus_fc_if.i.Size = sizeof(MLX4_FC_BUS_IF);
	p_fdo->bus_fc_if.i.Version = MLX4_BUS_FC_INTERFACE_VERSION;
	// Let the framework handle reference counting.
	p_fdo->bus_fc_if.i.InterfaceReference = WdfDeviceInterfaceReferenceNoOp;
	p_fdo->bus_fc_if.i.InterfaceDereference = WdfDeviceInterfaceDereferenceNoOp;

	p_fdo->bus_fc_if.pdev = &p_fdo->pci_dev;
	p_fdo->bus_fc_if.pmlx4_dev = to_mdev(p_fdo->pci_dev.ib_dev)->dev;
}

void __fix_registry(WDFDEVICE  Device, struct mlx4_dev_params *dev_params)
{
	NTSTATUS status = STATUS_SUCCESS;
	WDFKEY hKey = NULL;
	WDFKEY hParamsKey = NULL;
	PFDO_DEVICE_DATA p_fdo	= FdoGetData(Device);
	struct pci_dev *pdev = &p_fdo->pci_dev;    
	DECLARE_CONST_UNICODE_STRING(Parameters, L"Parameters");
	DECLARE_CONST_UNICODE_STRING(PortType, L"PortType");
	DECLARE_CONST_UNICODE_STRING(CapabilityFlag, L"CapabilityFlag");

#define  MAX_UVALUE 100
	WCHAR uvalue_data[MAX_UVALUE]={0};
	UNICODE_STRING uvalue;	  
	uvalue.Buffer = uvalue_data;
	uvalue.MaximumLength = MAX_UVALUE;
	uvalue.Length = 0;	  
		
	status = WdfDeviceOpenRegistryKey(Device,
									  PLUGPLAY_REGKEY_DRIVER, 
									  STANDARD_RIGHTS_ALL,
									  WDF_NO_OBJECT_ATTRIBUTES,
									  &hKey
									  );
	if(!NT_SUCCESS(status)) 
	{
		MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,
			("WdfDeviceOpenRegistryKey(\\Registry\\Machine\\Control\\Class\\...) Failed status = 0x%x\n", status));
		WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_WARN_REG_OPEN_DEV_KEY, 0, 0, 1,
			L"%#x", status );
		goto err;
	}

	status = WdfRegistryOpenKey(hKey,
								&Parameters,
								STANDARD_RIGHTS_ALL,
								WDF_NO_OBJECT_ATTRIBUTES,
								&hParamsKey
								);
	if(!NT_SUCCESS(status)) 
	{
		MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,("WdfRegistryOpenKey(Prameters) Failed status = 0x%x\n", status));
		WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_WARN_REG_ACTION, 0, 0, 3,
			L"%s", L"WdfRegistryOpenKey", L"%s", Parameters.Buffer, L"%#x", status );
		goto err;
	}

	if(pdev->dev->capability_flag.u.ver.fSinglePort)
	{
		if(dev_params->mod_port_type[0] == MLX4_PORT_TYPE_IB)
		{
			RtlStringCbCopyW(uvalue_data,MAX_UVALUE,L"ib,ib");
		}
		else if(dev_params->mod_port_type[0] == MLX4_PORT_TYPE_ETH)
		{
			RtlStringCbCopyW(uvalue_data,MAX_UVALUE,L"eth,eth");
		}					 
	}
	else
	{		 
		if(dev_params->mod_port_type[0] == MLX4_PORT_TYPE_IB && 
		   dev_params->mod_port_type[1] == MLX4_PORT_TYPE_IB
		   )
		{
			RtlStringCbCopyW(uvalue_data,MAX_UVALUE,L"ib,ib");
		}
		else if (dev_params->mod_port_type[0] == MLX4_PORT_TYPE_IB && 
				 dev_params->mod_port_type[1] == MLX4_PORT_TYPE_ETH
				 )
		{
			RtlStringCbCopyW(uvalue_data,MAX_UVALUE,L"ib,eth");
		}
		else if (dev_params->mod_port_type[0] == MLX4_PORT_TYPE_ETH && 
				 dev_params->mod_port_type[1] == MLX4_PORT_TYPE_IB
				 )
		{
			ASSERT(FALSE);
			RtlStringCbCopyW(uvalue_data,MAX_UVALUE,L"eth,ib");
		}
		else if (dev_params->mod_port_type[0] == MLX4_PORT_TYPE_ETH && 
				 dev_params->mod_port_type[1] == MLX4_PORT_TYPE_ETH
				 )
		{
			RtlStringCbCopyW(uvalue_data,MAX_UVALUE,L"eth,eth");
		}		 
	}

	
	uvalue.Length = (USHORT)(wcslen(uvalue_data)+1) * sizeof(WCHAR);
	
	status = WdfRegistryAssignValue(hParamsKey,
									&PortType,
									REG_SZ,
									uvalue.Length,
									uvalue.Buffer
									);		  
	if(!NT_SUCCESS(status)) 
	{
		MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,("WdfRegistryAssignValue(PortType) Failed status = 0x%x\n", status));
		WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_WARN_REG_ACTION, 0, 0, 3,
			L"%s", L"WdfRegistryAssignValue", L"%s", PortType.Buffer, L"%#x", status );
		goto err;
	}
	
	status = WdfRegistryAssignValue(hParamsKey,
									&CapabilityFlag,
									REG_DWORD,
									sizeof(pdev->dev->capability_flag),
									&pdev->dev->capability_flag.u.Value
									);		  
	if(!NT_SUCCESS(status)) 
	{
		MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,("WdfRegistryAssignValue(CapabilityFlag) Failed status = 0x%x\n", status));
		WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_WARN_REG_ACTION, 0, 0, 3,
			L"%s", L"WdfRegistryAssignValue", L"%s", CapabilityFlag.Buffer, L"%#x", status );
		goto err;
	}
	
err:	
	if (hKey != NULL) 
		WdfRegistryClose(hKey);

	if (hParamsKey != NULL) 
		WdfRegistryClose(hParamsKey);
}


static 
NTSTATUS
__start_card(
	IN WDFDEVICE  Device,
	IN PFDO_DEVICE_DATA p_fdo 
	)
{
		int i;
	int err;
	NTSTATUS status = STATUS_SUCCESS;
	struct pci_dev *pdev = &p_fdo->pci_dev;
	struct mlx4_dev_params			dev_params;
	struct mlx4_dev_params			dev_params_final;

	MLX4_ENTER(MLX4_DBG_DRV);

	if ( p_fdo->card_started )
		goto err; 

	status = __read_dev_params(Device, &dev_params);
	if( !NT_SUCCESS( status ) ) {		
		MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,
			("__read_dev_params Failed status = 0x%x\n", status));
		goto err;
	}

	// enable the card
	status = pci_hca_enable( &pdev->bus_pci_ifc, &pdev->pci_cfg_space );
	if( !NT_SUCCESS( status ) ) {
		MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,
			("pci_hca_enable Failed status = 0x%x\n", status));
		goto err;
	}

	//
	// init the card
	//

#ifndef USE_WDM_INTERRUPTS
	// enable interrupts for start up
	for ( i = 0; i < MLX4_MAX_INTERRUPTS; ++i ) 
		WdfInterruptEnable(p_fdo->interrupt[i].WdfInterrupt);
#endif	

	// NET library
	err = mlx4_init_one( &p_fdo->pci_dev, &dev_params );
	if (err) {
		status = errno_to_ntstatus(err);		
		MLX4_PRINT( TRACE_LEVEL_ERROR  ,MLX4_DBG_DEV  ,
			("mlx4_init_one Failed status = 0x%x\n", status));
		goto err;
	}

	for (i = 1; i <= pdev->dev->caps.num_ports; ++i) 
	{
		dev_params_final.mod_port_type[i-1] = pdev->dev->caps.port_type[i];
	}
	
	__fix_registry(Device,&dev_params_final);

#ifndef USE_WDM_INTERRUPTS
	//
	// complete filling interrupt context (for more efficiency)
	//
	for ( i = 0; i < MLX4_MAX_INTERRUPTS; ++i ) {
		struct mlx4_priv *priv = mlx4_priv( p_fdo->pci_dev.dev );
		PINTERRUPT_DATA p_isr_ctx = WdfObjectGetTypedContext( 
			p_fdo->interrupt[i].WdfInterrupt, INTERRUPT_DATA );

		p_isr_ctx->eq = &priv->eq_table.eq[i];
	}
#endif

	InitInterface(p_fdo);

	p_fdo->card_started = TRUE;

err:
	MLX4_EXIT( MLX4_DBG_DRV );
	return status;
}

static
void
__stop_card(
	IN PFDO_DEVICE_DATA p_fdo
	)
{
	if ( p_fdo->card_started ) {
		p_fdo->card_started = FALSE;
		mlx4_remove_one( &p_fdo->pci_dev);
	}
}

NTSTATUS
EvtDeviceD0Entry(
	IN WDFDEVICE  Device,
	IN WDF_POWER_DEVICE_STATE  PreviousState
	)
{
	NTSTATUS status = STATUS_SUCCESS;
	PFDO_DEVICE_DATA p_fdo	= FdoGetData(Device);
	struct pci_dev *pdev	= &p_fdo->pci_dev;
	struct mlx4_dev *mdev;
	struct ib_device_attr props;
	struct ib_device *p_ibdev;

	MLX4_ENTER(MLX4_DBG_DRV);

	MLX4_PRINT(TRACE_LEVEL_INFORMATION, MLX4_DBG_DRV, ("PreviousState 0x%x\n", PreviousState));

	// start card (needed after Hibernation or standby)
	if ( !pdev->start_event_taken )
		KeWaitForSingleObject(&g.start_event, Executive, KernelMode, FALSE, NULL);
	if (PreviousState > WdfPowerDeviceD0)
		status = __start_card( Device, p_fdo );
	pdev->start_event_taken = FALSE;
	KeSetEvent(&g.start_event, 0, FALSE);
	if ( !NT_SUCCESS( status ) ) 
		goto err;
	mdev = pdev->dev;

	// create child device
	if ( !(g.mode_flags & MLX4_MODE_SKIP_CHILDS) ) {
		status = __do_static_enumeration(Device);
			if (!NT_SUCCESS(status)) {
				MLX4_PRINT_EV(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, ("%s: __do_static_enumeration failed with 0x%x\n", 
					mdev->pdev->name, status));
				goto err;
			}
	}

	// Log Success Message
	MLX4_PRINT(TRACE_LEVEL_INFORMATION ,MLX4_DBG_DRV ,
		("Ven %x Dev %d Fw %d.%d.%d\n", 
		(unsigned)pdev->ven_id, (unsigned)pdev->dev_id,
		(int) (mdev->caps.fw_ver >> 32),
		(int) (mdev->caps.fw_ver >> 16) & 0xffff, 
		(int) (mdev->caps.fw_ver & 0xffff)
		));
	memset ( &props, 0, sizeof( props) );
	p_ibdev = pdev->ib_dev;
	(p_ibdev->query_device)( p_ibdev, &props );
    if (!mlx4_is_mfunc(mdev) || mlx4_is_master(mdev))
    {
    	WriteEventLogEntryData( pdev->p_self_do, (ULONG)EVENT_MLX4_INFO_DEV_STARTED, 0, 0, 16,
    		L"%S", pdev->name, 
    		L"%04x", (ULONG)pdev->ven_id, 
    		L"%04x", (ULONG)pdev->dev_id,
    		L"%04x", (ULONG)pdev->sub_vendor_id,
    		L"%04x", (ULONG)pdev->sub_system_id,
    		L"%02x", (ULONG)pdev->revision_id,
    		L"%d", (int) (mdev->caps.fw_ver >> 32),
    		L"%d", (int) (mdev->caps.fw_ver >> 16) & 0xffff, 
    		L"%d", (int) (mdev->caps.fw_ver & 0xffff),
    		L"%08x", be32_to_cpu(*(PULONG)((PUCHAR)&p_ibdev->node_guid + 0)), 
    		L"%08x", be32_to_cpu(*(PULONG)((PUCHAR)&p_ibdev->node_guid + 4)), 
    		L"%d", mdev->caps.num_ports,
    		L"%s", mdev->dev_params.mod_port_type[0] == MLX4_PORT_TYPE_IB ? L"IB" : L"ETH",
    		L"%s", mdev->dev_params.mod_port_type[1] == MLX4_PORT_TYPE_IB ? L"IB" : L"ETH",
    		L"%s", mdev->caps.port_type[1] == MLX4_PORT_TYPE_IB ? L"IB" : L"ETH"
    		L"%s", mdev->caps.port_type[2] == MLX4_PORT_TYPE_IB ? L"IB" : L"ETH"
    		); 
    }
err:
	MLX4_EXIT( MLX4_DBG_DRV );
	return status;
}

NTSTATUS
EvtDeviceD0Exit(
	IN WDFDEVICE  Device,
	IN WDF_POWER_DEVICE_STATE  TargetState
	)
{
	PFDO_DEVICE_DATA p_fdo	= FdoGetData(Device);
	
	MLX4_ENTER(MLX4_DBG_DRV);
	MLX4_PRINT(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, ("TargetState 0x%x\n", TargetState));

	if (TargetState > WdfPowerDeviceD0)
		__stop_card( p_fdo );

	MLX4_EXIT( MLX4_DBG_DRV );

	return STATUS_SUCCESS;
}


char *GetNotificaationTypeString(WDF_SPECIAL_FILE_TYPE NotificationType)
{
	switch(NotificationType) {
		case WdfSpecialFilePaging : return "WdfSpecialFilePaging";
		case WdfSpecialFileHibernation : return "WdfSpecialFileHibernation";
		case WdfSpecialFileDump : return "WdfSpecialFileDump";
		default:
			return "Unknown type";
	}

}

VOID
EvtDeviceUsageNotification (
	IN WDFDEVICE  Device,
	IN WDF_SPECIAL_FILE_TYPE NotificationType,
	IN BOOLEAN	IsInNotificationPath
	)
{
	UNUSED_PARAM(Device);
	
	MLX4_PRINT(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, 
		("EvtDeviceUsageNotification called NotificationType = %s ,IsInNotificationPath=%s \n",
		GetNotificaationTypeString(NotificationType), 
		IsInNotificationPath ? "true" : "false"));
	// BUGBUG: Tzachid 27/4/2010 Need to implment this functions as part of the iSCSI install.
	// For more details see IRP_MN_DEVICE_USAGE_NOTIFICATION in the wdk

}



/* Forwards the request to the HCA's PDO. */
static 
void
__put_bus_ifc(
		IN	BUS_INTERFACE_STANDARD		*pBusIfc )
{
	MLX4_ENTER(MLX4_DBG_DRV);
	MLX4_PRINT(TRACE_LEVEL_INFORMATION, MLX4_DBG_DRV, ("pBusIfc=0x%p\n", pBusIfc));
	pBusIfc->InterfaceDereference( pBusIfc->Context );
	MLX4_EXIT( MLX4_DBG_DRV );
}

static 
NTSTATUS
__get_bus_ifc(
	IN				PFDO_DEVICE_DATA const	p_fdo,
	IN		const	GUID* const				pGuid,
	OUT			BUS_INTERFACE_STANDARD	*pBusIfc )
{
	NTSTATUS status;
	WDFDEVICE  FdoDevice = p_fdo->FdoDevice;
	MLX4_ENTER(MLX4_DBG_DRV);
	
	status = WdfFdoQueryForInterface( FdoDevice, pGuid, (PINTERFACE)pBusIfc,
		sizeof(BUS_INTERFACE_STANDARD), 1, NULL );
	MLX4_EXIT( MLX4_DBG_DRV );
	return status;
}

static
void
__put_dma_adapter(
	IN PFDO_DEVICE_DATA p_fdo,
	IN PDMA_ADAPTER 	p_dma )
{
	UNUSED_PARAM(p_fdo);
	UNUSED_PARAM(p_dma);
	MLX4_ENTER(MLX4_DBG_DRV);
	MLX4_EXIT( MLX4_DBG_DRV );
}


// this routine releases the resources, taken in __get_resources
static
void 
__put_resources(
	IN PFDO_DEVICE_DATA p_fdo
	)
{
	struct pci_dev *pdev = &p_fdo->pci_dev;

	MLX4_ENTER(MLX4_DBG_DRV);

	if (pdev->msix_info.valid) 
		pci_free_msix_info_resources(&pdev->msix_info);

	if (p_fdo->dma_adapter_taken) {
		p_fdo->dma_adapter_taken = FALSE;
		__put_dma_adapter( p_fdo, pdev->p_dma_adapter );
	}

	if (p_fdo->pci_bus_ifc_taken) {
		p_fdo->pci_bus_ifc_taken = FALSE;
		__put_bus_ifc(&pdev->bus_pci_ifc);
	}

	if (pdev->p_msix_map)
		kfree(pdev->p_msix_map);

	
	MLX4_EXIT( MLX4_DBG_DRV );
}

static
NTSTATUS 
__get_dma_adapter(
	IN PFDO_DEVICE_DATA p_fdo,
	OUT PDMA_ADAPTER *	pp_dma )
{
	NTSTATUS status;
	WDF_DMA_ENABLER_CONFIG	dmaConfig;
	
	MLX4_ENTER(MLX4_DBG_DRV);

	WDF_DMA_ENABLER_CONFIG_INIT( &dmaConfig,
		WdfDmaProfileScatterGather64, 0x80000000 - 1 );

	status = WdfDmaEnablerCreate( p_fdo->FdoDevice,
		&dmaConfig, WDF_NO_OBJECT_ATTRIBUTES, &p_fdo->dma_enabler );
	if (!NT_SUCCESS (status)) {
		return status;
	}
	
	*pp_dma = WdfDmaEnablerWdmGetDmaAdapter( 
		p_fdo->dma_enabler, WdfDmaDirectionReadFromDevice );

	MLX4_EXIT( MLX4_DBG_DRV );
	return status;
}

// this routine fills pci_dev structure, containing all HW 
// and some other necessary common resources
static
NTSTATUS 
__get_resources(
	IN PFDO_DEVICE_DATA p_fdo,
	IN WDFCMRESLIST  ResourcesRaw,
	IN WDFCMRESLIST  ResourcesTranslated
	)
{
	NTSTATUS status;
	ULONG i, k=0;
	PCM_PARTIAL_RESOURCE_DESCRIPTOR  desc;
	PCM_PARTIAL_RESOURCE_DESCRIPTOR  desc_raw;
	BUS_INTERFACE_STANDARD	bus_pci_ifc;
	struct pci_dev *pdev = &p_fdo->pci_dev;

	MLX4_ENTER(MLX4_DBG_DRV);

	//
	// Get PCI BUS interface
	// 
	status = __get_bus_ifc( p_fdo, &GUID_BUS_INTERFACE_STANDARD, &bus_pci_ifc );
	if( !NT_SUCCESS( status ) ) {
		MLX4_PRINT(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, 
			("failed: status=0x%x\n", status));
		return status;
	}
	RtlCopyMemory( &pdev->bus_pci_ifc, &bus_pci_ifc, sizeof(BUS_INTERFACE_STANDARD) );
	p_fdo->pci_bus_ifc_taken = TRUE;

	// 
	// get HW resources
	//
	for (i = 0; i < WdfCmResourceListGetCount(ResourcesTranslated); i++) {

		desc = WdfCmResourceListGetDescriptor( ResourcesTranslated, i );
		desc_raw = WdfCmResourceListGetDescriptor( ResourcesRaw, i );

		switch (desc->Type) {

			case CmResourceTypeMemory:
				MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV,
					("EvtPrepareHardware(Raw): Desc %d: Memory: Start %#I64x, Length %#x\n", 
					i, desc_raw->u.Memory.Start.QuadPart, desc_raw->u.Memory.Length ));
				MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV,
					("EvtPrepareHardware: Desc %d: Memory: Start %#I64x, Length %#x\n", 
					i, desc->u.Memory.Start.QuadPart, desc->u.Memory.Length ));

				if (k < N_BARS) {
					pdev->bar[k].phys = desc->u.Memory.Start.QuadPart;
					pdev->bar[k].size = (SIZE_T)desc->u.Memory.Length;
				}
				k++;
				break;

#ifdef USE_WDM_INTERRUPTS
			case CmResourceTypeInterrupt:
				if (!pdev->n_msi_vectors_alloc)
					pdev->int_info = *desc;
				if (desc->Flags & CM_RESOURCE_INTERRUPT_MESSAGE) {
					pdev->n_msi_vectors_alloc = (u8)(pdev->n_msi_vectors_alloc+desc_raw->u.MessageInterrupt.Raw.MessageCount);
					MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV,
						("EvtPrepareHardware: Desc %d: MsiInterrupt: Share %d, Flags %#x, Level %d, Vector %#x, Affinity %#x\n", 
						i, desc->ShareDisposition, desc->Flags,
						desc->u.MessageInterrupt.Translated.Level, 
						desc->u.MessageInterrupt.Translated.Vector, 
						(u32)desc->u.MessageInterrupt.Translated.Affinity ));
					MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV,
						("EvtPrepareHardware: Desc %d: RawMsiInterrupt: Share %d, Flags %#x, MessageCount %#hx, Vector %#x, Affinity %#x\n", 
						i, desc_raw->ShareDisposition, desc_raw->Flags,
						desc_raw->u.MessageInterrupt.Raw.MessageCount, 
						desc_raw->u.MessageInterrupt.Raw.Vector,
						(u32)desc_raw->u.MessageInterrupt.Raw.Affinity ));
				}
				else { // line-based interrupt
					MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV,
						("EvtPrepareHardware: Desc %d: LineInterrupt: Share %d, Flags %#x, Level %d, Vector %#x, Affinity %#x\n", 
						i, desc->ShareDisposition, desc->Flags,
						desc->u.Interrupt.Level, desc->u.Interrupt.Vector, 
						(u32)desc->u.Interrupt.Affinity ));
				}
				break;
#endif

			default:
				//
				// Ignore all other descriptors.
				//
				break;
		}
	}
	if (i ==0) {
		// This means that no resources are found
		MLX4_PRINT(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, ("WdfCmResourceListGetCount: returned 0, quiting\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	//
	// get uplink info. 
	//
	status = pci_save_config( &pdev->bus_pci_ifc, 
		&pdev->pci_cfg_space );
	if( !NT_SUCCESS( status ) )
	{
		MLX4_PRINT(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, 
			("Failed to save HCA config: status=0x%x\n", status));
		goto err;
	}
	pci_get_uplink_info( pdev, &pdev->pci_cfg_space, &pdev->uplink_info );

	//
	// allocate DMA adapter
	//
	status = __get_dma_adapter( p_fdo, &pdev->p_dma_adapter );
	if( !NT_SUCCESS( status ) )
	{
		MLX4_PRINT(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, 
			("Failed to get DMA adapter: status=0x%x\n", status));
		goto err;
	}
	p_fdo->dma_adapter_taken = TRUE;

	//
	// allocate MSI-X vector map table
	//
	if ( pdev->n_msi_vectors_alloc )
	{
		pdev->p_msix_map = kzalloc(sizeof(struct msix_map) * pdev->n_msi_vectors_alloc, GFP_KERNEL);
		if ( !pdev->p_msix_map )
		{
			MLX4_PRINT(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, 
				("Failed to allocate MSI-X vector map table\n"));
			goto err;
		}
	}
	
	//
	// fill more fields in pci_dev
	//
	pdev->ven_id = pdev->pci_cfg_space.VendorID;
	pdev->dev_id = pdev->pci_cfg_space.DeviceID;
	pdev->sub_vendor_id = pdev->pci_cfg_space.u.type0.SubVendorID;
	pdev->sub_system_id = pdev->pci_cfg_space.u.type0.SubSystemID;
	pdev->revision_id = pdev->pci_cfg_space.RevisionID;
	pdev->p_self_do = WdfDeviceWdmGetDeviceObject(p_fdo->FdoDevice);
	pdev->pdo = WdfDeviceWdmGetPhysicalDevice(p_fdo->FdoDevice);
	pdev->p_stat->pdo = pdev->pdo;
	
	MLX4_EXIT( MLX4_DBG_DRV );
	return STATUS_SUCCESS;
err:
	__put_resources(p_fdo);
	MLX4_EXIT( MLX4_DBG_DRV );
	return status;
}

NTSTATUS
EvtPrepareHardware(
	IN WDFDEVICE  Device,
	IN WDFCMRESLIST  ResourcesRaw,
	IN WDFCMRESLIST  ResourcesTranslated
	)
{
	NTSTATUS status;
	PFDO_DEVICE_DATA p_fdo	= FdoGetData(Device);
	struct pci_dev *pdev = &p_fdo->pci_dev;
	struct mlx4_dev *mdev;
	WDFMEMORY  memory;
	WDF_OBJECT_ATTRIBUTES  attributes;
    
	MLX4_ENTER(MLX4_DBG_DRV);

	KeWaitForSingleObject(&g.start_event, Executive, KernelMode, FALSE, NULL);
	pdev->start_event_taken = TRUE;

	// get resources
	status = __get_resources( p_fdo, ResourcesRaw, ResourcesTranslated );
	if( !NT_SUCCESS( status ) ) {
		MLX4_PRINT(TRACE_LEVEL_ERROR, MLX4_DBG_DRV, ("__get_resources failed: status=0x%x\n", status));
		goto err;
	}

	// get card location
	p_fdo->pci_dev.location[0] = 0;
	WDF_OBJECT_ATTRIBUTES_INIT(&attributes);    
	attributes.ParentObject = Device;

	
	status = WdfDeviceAllocAndQueryProperty(Device,
											DevicePropertyBusNumber,
											NonPagedPool,
											&attributes,
											&memory
											);	  
	if(NT_SUCCESS(status))
	{
		UCHAR *ptr;
		// get bus number
		ptr = WdfMemoryGetBuffer(memory, NULL);
		p_fdo->pci_dev.pci_bus = *(PULONG)ptr;
		WdfObjectDelete(memory);
	}
	else
	{
		WriteEventLogEntryData( p_fdo->pci_dev.p_self_do,
								(ULONG)EVENT_MLX4_ERROR_GET_LOCATION, 0, 0, 1,
								L"%x", status);
	}

	status = WdfDeviceAllocAndQueryProperty(Device,
											DevicePropertyAddress,
											NonPagedPool,
											&attributes,
											&memory
											);
	if(NT_SUCCESS(status)) 
	{
		UCHAR *ptr;
		// get function number
		ptr = WdfMemoryGetBuffer(memory, NULL);
		p_fdo->pci_dev.pci_func = *(PULONG)ptr;

		WdfObjectDelete(memory);
	}
	else
	{
		WriteEventLogEntryData( p_fdo->pci_dev.p_self_do,
								(ULONG)EVENT_MLX4_ERROR_GET_LOCATION, 0, 0, 1,
								L"%x", status);
	}
	
	p_fdo->pci_dev.pci_device = 0;

	status = RtlStringCbPrintfW( (LPWSTR)p_fdo->pci_dev.location, 
			sizeof(p_fdo->pci_dev.location), L"PCI bus %d, device %d, function %d",
			p_fdo->pci_dev.pci_bus,p_fdo->pci_dev.pci_device,p_fdo->pci_dev.pci_func );
	ASSERT(NT_SUCCESS( status ));
	
	p_fdo->pci_dev.devfn = (u16)PCI_DEVFN(p_fdo->pci_dev.pci_device, p_fdo->pci_dev.pci_func);
	RtlStringCbPrintfA( p_fdo->pci_dev.name, sizeof(p_fdo->pci_dev.name),
		"MLX4_SF_%d_%d_%d", p_fdo->pci_dev.pci_bus, 
		p_fdo->pci_dev.pci_device, p_fdo->pci_dev.pci_func );

	pdev->p_wdf_device = Device;
	pdev->ib_hca_created = 0;

	// start the card
	status = __start_card(Device, p_fdo );
	if( !NT_SUCCESS( status ) ) 
		goto err;
	mdev = pdev->dev;

	// get VPD - as far as it just info, we proceed the work in case of error
	status = pci_get_vpd( &pdev->bus_pci_ifc, &pdev->pci_cfg_space, &pdev->vpd, &pdev->vpd_size );
	if( !NT_SUCCESS( status ) )
	{
		MLX4_PRINT(TRACE_LEVEL_WARNING, MLX4_DBG_DRV, 
			("%s: Failed to read VPD from the card (status=0x%x). Continue the work.\n", 
			mdev->pdev->name, status));
		status = 0;
	}
err:
	if( !NT_SUCCESS( status ) )
		st_dev_rmv( p_fdo->pci_dev.p_stat );
	MLX4_EXIT( MLX4_DBG_DRV );
	return status;
}

void fix_bus_ifc(struct pci_dev *pdev)
{
	PFDO_DEVICE_DATA p_fdo;

	p_fdo =  CONTAINING_RECORD(pdev, FDO_DEVICE_DATA, pci_dev);
	p_fdo->bus_ib_ifc.p_ibdev = p_fdo->pci_dev.ib_dev;
	p_fdo->bus_ib_ifc.pmlx4_dev = to_mdev(p_fdo->pci_dev.ib_dev)->dev;

	p_fdo->bus_fc_if.pmlx4_dev = to_mdev(p_fdo->pci_dev.ib_dev)->dev;
}

NTSTATUS
EvtReleaseHardware(
	IN WDFDEVICE  Device,
	IN WDFCMRESLIST  ResourcesTranslated
	)
{
	PFDO_DEVICE_DATA p_fdo	= FdoGetData(Device);
	struct pci_dev *pdev = &p_fdo->pci_dev;
	int i;

	UNUSED_PARAM(ResourcesTranslated);

	MLX4_ENTER(MLX4_DBG_DRV);

	if ( pdev->start_event_taken ) {
		pdev->start_event_taken = FALSE;
		KeSetEvent(&g.start_event, 0, FALSE);
	}

	if ( pdev->vpd ) {
		kfree( pdev->vpd );
		pdev->vpd = NULL;
		pdev->vpd_size = 0;
	}
	__stop_card( p_fdo );
	__put_resources( p_fdo );
	for (i=0; i<MAX_HCA_CARDS; ++i) {
		if ( g.p_fdo[i] == p_fdo) {
			g.p_fdo[i] = NULL;
			break;
		}
	}
	st_dev_rmv( p_fdo->pci_dev.p_stat );

	MLX4_EXIT( MLX4_DBG_DRV );
	return STATUS_SUCCESS;
}

#ifndef USE_WDM_INTERRUPTS

static
NTSTATUS 
__create_interrupt(
	IN WDFDEVICE				device,
	IN int						int_num,
	IN PFN_WDF_INTERRUPT_ISR	isr,
	IN PFN_WDF_INTERRUPT_DPC	dpc,
	IN PFDO_DEVICE_DATA			p_fdo,
	OUT WDFINTERRUPT		*	p_int_obj
	)
{
	NTSTATUS Status;

	WDF_INTERRUPT_CONFIG  interruptConfig;
	WDF_OBJECT_ATTRIBUTES  interruptAttributes;
	PINTERRUPT_DATA p_isr_ctx;

	MLX4_ENTER(MLX4_DBG_DRV);

	WDF_INTERRUPT_CONFIG_INIT( &interruptConfig, isr, dpc );
	
	interruptConfig.EvtInterruptEnable = EvtEnableInterrupt;
	interruptConfig.EvtInterruptDisable = EvtDisableInterrupt;
	
	
	WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE( 
		&interruptAttributes, INTERRUPT_DATA );

	Status = WdfInterruptCreate( device,
		&interruptConfig, &interruptAttributes, p_int_obj );

	p_isr_ctx = WdfObjectGetTypedContext( *p_int_obj, INTERRUPT_DATA );
	p_isr_ctx->int_num = int_num;
	p_isr_ctx->p_fdo = p_fdo;
	p_isr_ctx->eq = NULL;

	// one can call WdfInterruptSetPolicy() to set the policy, affinity etc

	MLX4_EXIT( MLX4_DBG_DRV );
	return Status;
}

#endif

inline void InitBusIsr(
	 struct VipBusIfc* pVipBusIfc
	)
{
	memset(pVipBusIfc, 0, sizeof(struct VipBusIfc));
	KeInitializeEvent(&pVipBusIfc->NicData.ConfigChangeEvent, SynchronizationEvent, TRUE);
}


//#define DONT_START_ON_BOOT

NTSTATUS
EvtDriverDeviceAdd(
	IN WDFDRIVER		Driver,
	IN PWDFDEVICE_INIT	DeviceInit
	)
/*++
Routine Description:

	EvtDriverDeviceAdd is called by the framework in response to AddDevice
	call from the PnP manager. We create and initialize a device object to
	represent a new instance of mxe bus.

Arguments:

	Driver - Handle to a framework driver object created in DriverEntry

	DeviceInit - Pointer to a framework-allocated WDFDEVICE_INIT structure.

Return Value:

	NTSTATUS

--*/
{
#ifndef USE_WDM_INTERRUPTS
	int i;
#endif
	WDF_OBJECT_ATTRIBUTES			attributes;
	NTSTATUS						status;
	WDFDEVICE						device;
	PFDO_DEVICE_DATA				p_fdo;
	PNP_BUS_INFORMATION				busInfo;
	WDF_PNPPOWER_EVENT_CALLBACKS	Callbacks;
	WDF_DEVICE_PNP_CAPABILITIES		pnpCaps;
	int i;
	WDFMEMORY  memory;
	int bus = 0, dev = 0, function= 0;
	UCHAR *ptr;
	

	UNREFERENCED_PARAMETER(Driver);

	PAGED_CODE ();
	MLX4_ENTER(MLX4_DBG_DRV);

#ifdef DONT_START_ON_BOOT
	{
		int    QueryTimeIncrement = KeQueryTimeIncrement();
		LARGE_INTEGER	Ticks;

		KeQueryTickCount(&Ticks);
		if (Ticks.QuadPart * QueryTimeIncrement / 10000 < 30000) // 10,000 moves from 100ns to ms
		{
			return STATUS_NO_MEMORY;
		}
	}
#endif
	//
	// register PnP & Power stuff
	//
	WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&Callbacks);
	Callbacks.EvtDevicePrepareHardware = EvtPrepareHardware;
	Callbacks.EvtDeviceReleaseHardware = EvtReleaseHardware;
	Callbacks.EvtDeviceD0Entry = EvtDeviceD0Entry;
	Callbacks.EvtDeviceD0Exit = EvtDeviceD0Exit;
	Callbacks.EvtDeviceUsageNotification = EvtDeviceUsageNotification;

	WdfDeviceInitSetPnpPowerEventCallbacks( DeviceInit, &Callbacks );
	
	//
	// Initialize all the properties specific to the device.
	// Framework has default values for the one that are not
	// set explicitly here. So please read the doc and make sure
	// you are okay with the defaults.
	//
	WdfDeviceInitSetDeviceType(DeviceInit, FILE_DEVICE_BUS_EXTENDER);
	WdfDeviceInitSetExclusive(DeviceInit, TRUE);

	// 
	// Initialize attributes structure to specify size and accessor function
	// for storing device context.
	//
	WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attributes, FDO_DEVICE_DATA);

	//
	// Create a framework device object. In response to this call, framework
	// creates a WDM deviceobject.
	//
	status = WdfDeviceCreate(&DeviceInit, &attributes, &device);
	if (!NT_SUCCESS(status)) {
		MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV, 
			("WdfDeviceCreate failed with 0x%x\n", status));
		goto end;
	}

	//
	// set PnP capabilities
	//
	WDF_DEVICE_PNP_CAPABILITIES_INIT(&pnpCaps);
	pnpCaps.Removable = WdfTrue;
	pnpCaps.SurpriseRemovalOK = WdfTrue;
	
	WdfDeviceSetPnpCapabilities( device, &pnpCaps );

	//
	// Get the device context.
	//
	p_fdo = FdoGetData(device);
	RtlZeroMemory(p_fdo, sizeof(FDO_DEVICE_DATA));
	p_fdo->FdoDevice = device; 
	for (i=0; i<MAX_HCA_CARDS; ++i) {
		if (!g.p_fdo[i]) {
			g.p_fdo[i] = p_fdo;
			break;
		}
	}

	//
	// Init the BusIsr data
	//
	InitBusIsr(&p_fdo->mtnic_Ifc);

	//
	// Purpose of this lock is documented in PlugInDevice routine below.
	//
	WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
	attributes.ParentObject = device;
	status = WdfWaitLockCreate(&attributes, &p_fdo->ChildLock);
	if (!NT_SUCCESS(status)) {
		MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV, 
			("WdfWaitLockCreate failed with 0x%x\n", status));
		goto end;
	}

	//
	// This value is used in responding to the IRP_MN_QUERY_BUS_INFORMATION
	// for the child devices. This is an optional information provided to
	// uniquely identify the bus the device is connected.
	//

	WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
	attributes.ParentObject = device;
    
	status = WdfDeviceAllocAndQueryProperty(device,
											DevicePropertyBusNumber,
											NonPagedPool,
											&attributes,
											&memory
											);	  
	if(!NT_SUCCESS(status)) {
		MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV, 
			("WdfDeviceAllocAndQueryProperty failed with 0x%x\n", status));
		goto end;
	}


	// get bus number
	ptr = WdfMemoryGetBuffer(memory, NULL);
	bus = *(PULONG)ptr;

	WdfObjectDelete(memory);
	

	status = WdfDeviceAllocAndQueryProperty(device,
											DevicePropertyAddress,
											NonPagedPool,
											&attributes,
											&memory
											);

	if(!NT_SUCCESS(status)) {
		MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV, 
			("WdfDeviceAllocAndQueryProperty failed with 0x%x\n", status));	
		goto end;
	}


	// get function number
	ptr = WdfMemoryGetBuffer(memory, NULL);
	function = *(PULONG)ptr;

	WdfObjectDelete(memory);
	

	//
	//	We don't use dev number in code
	//	Since in all machines the value is zero 
	//	we will set it to zero
	//
	dev = 0;

	
	busInfo.BusTypeGuid = GUID_BUS_TYPE_INTERNAL;
	busInfo.LegacyBusType = Internal;
	busInfo.BusNumber = bus;

	WdfDeviceSetBusInformationForChildren(device, &busInfo);
	WdfDeviceSetSpecialFileSupport(device, WdfSpecialFilePaging, TRUE);
	WdfDeviceSetSpecialFileSupport(device, WdfSpecialFileHibernation, TRUE);
	WdfDeviceSetSpecialFileSupport(device, WdfSpecialFileDump, TRUE);

	//
	// WMI
	//
	status = WmiRegistration(device);
	if (!NT_SUCCESS(status)) {
		MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV, 
			("WmiRegistration failed with 0x%x\n", status));
		goto end;
	}

#ifndef USE_WDM_INTERRUPTS

	//
	// create interrupt objects
	//
	for ( i = 0; i < MLX4_MAX_INTERRUPTS; ++i ) {
		status = __create_interrupt( p_fdo->FdoDevice, i, EvtInterruptIsr,
			NULL, p_fdo, &p_fdo->interrupt[i].WdfInterrupt );
		if (NT_SUCCESS(status)) 
			p_fdo->interrupt[i].valid = TRUE;
		else {
			MLX4_PRINT(TRACE_LEVEL_ERROR, MLX4_DBG_DRV,
				("WdfInterruptCreate failed %#x\n", status ));
			goto end;
		}
	}

#endif

	// init pdev lock
	KeInitializeEvent( &p_fdo->pci_dev.remove_dev_lock, SynchronizationEvent , TRUE);
	
	// statistics
	p_fdo->pci_dev.p_stat = st_dev_add(bus, function);
	if ( p_fdo->pci_dev.p_stat ) {
		p_fdo->pci_dev.p_stat->p_fdo = p_fdo;
		p_fdo->pci_dev.p_stat->h_wdf_device = device;
		p_fdo->pci_dev.p_stat->flags = g.stat_flags;
		status = STATUS_SUCCESS;
	} else {
		MLX4_PRINT(TRACE_LEVEL_ERROR, MLX4_DBG_DRV,("st_dev_add failed (bus %d, func %d) \n" ,
			bus, function ));
		status = STATUS_NO_MEMORY;
	}
	

end:	
	MLX4_EXIT( MLX4_DBG_DRV );
	return status;
}



void
EvtDriverUnload(
	IN		WDFDRIVER  Driver
	)
{
	MLX4_ENTER( MLX4_DBG_DRV );

	UNUSED_PARAM( Driver );

	mlx4_ib_cleanup();
	core_cleanup();

	MLX4_EXIT( MLX4_DBG_DRV );
#if defined(EVENT_TRACING)
	WPP_CLEANUP(WdfDriverWdmGetDriverObject(Driver));
#endif

}

static
NTSTATUS
__read_registry(WDFDRIVER *hDriver)
{
	DECLARE_CONST_UNICODE_STRING(debugLevel, L"DebugLevel");
	DECLARE_CONST_UNICODE_STRING(debugFlags, L"DebugFlags");

	// "log maximum number of QPs per HCA"
	DECLARE_CONST_UNICODE_STRING(numQp, L"LogNumQp");

	// "log number of RDMARC buffers per QP"
	DECLARE_CONST_UNICODE_STRING(numRdmaRc, L"LogNumRdmaRc");

	// "log maximum number of SRQs per HCA"
	DECLARE_CONST_UNICODE_STRING(numSrq, L"LogNumSrq");

	// "log maximum number of CQs per HCA"
	DECLARE_CONST_UNICODE_STRING(numCq, L"LogNumCq");

	// "log maximum number of multicast groups per HCA"
	DECLARE_CONST_UNICODE_STRING(numMcg, L"LogNumMcg");

	// "log maximum number of memory protection table entries per HCA"
	DECLARE_CONST_UNICODE_STRING(numMpt, L"LogNumMpt");

	// "log maximum number of memory translation table segments per HCA"
	DECLARE_CONST_UNICODE_STRING(numMtt, L"LogNumMtt");	

	// "Maximum number of MACs per ETH port (1-127, default 0"
	DECLARE_CONST_UNICODE_STRING(numMac, L"LogNumMac");	

	// "Maximum number of VLANs per ETH port (0-126, default 0)"
	DECLARE_CONST_UNICODE_STRING(numVlan, L"LogNumVlan");	

	// "Enable steering by VLAN priority on ETH ports (0/1, default 0)"
	DECLARE_CONST_UNICODE_STRING(usePrio, L"UsePrio");	

	// "max number of FC_EXCH (0-N, default 0)"
	DECLARE_CONST_UNICODE_STRING(numFcExch, L"NumFcExch");	

	// "Enable Quality of Service support in the HCA if > 0, (default 1)"
	DECLARE_CONST_UNICODE_STRING(enableQoS, L"EnableQoS");	

	// "Block multicast loopback packets if > 0 (default 1)"
	DECLARE_CONST_UNICODE_STRING(BlockMcastLB, L"BlockMcastLoopBack");	

	// "Measure the interrupt from the first packet (default 1)"
	DECLARE_CONST_UNICODE_STRING(InterruptFromFirstPacket, L"InterruptFromFirstPacket");	

	// "ProcessorAffinity"
	DECLARE_CONST_UNICODE_STRING(ProcessorAffinity, L"ProcessorAffinity");

	// "Stat Flags"
	DECLARE_CONST_UNICODE_STRING(StatFlags, L"StatFlags");

	// "Work Mode Flags"
	DECLARE_CONST_UNICODE_STRING(ModeFlags, L"ModeFlags");

	// "Slave Number"
	DECLARE_CONST_UNICODE_STRING(SlaveNum, L"SlaveNum");

	// "LogMttsPerSeg"
	DECLARE_CONST_UNICODE_STRING(LogMttsPerSeg, L"LogMttsPerSeg");

	// "SrIov"
	DECLARE_CONST_UNICODE_STRING(SrIov, L"SrIov");

	// "ProbeVf"
	DECLARE_CONST_UNICODE_STRING(ProbeVf, L"ProbeVf");

	// "Set4kMtu"
	DECLARE_CONST_UNICODE_STRING(Set4kMtu, L"Set4kMtu");

	// "EnableFcPreT11"
	DECLARE_CONST_UNICODE_STRING(EnableFcPreT11, L"EnableFcPreT11");

	ULONG value;
	WDFKEY hKey = NULL;
	NTSTATUS status = STATUS_SUCCESS;
	
	status = WdfDriverOpenParametersRegistryKey( *hDriver,
		STANDARD_RIGHTS_ALL, WDF_NO_OBJECT_ATTRIBUTES, &hKey );

	if (NT_SUCCESS (status)) {

		//
		// Read general values
		//
		status = WdfRegistryQueryULong(hKey, &debugLevel, &value);
		if (NT_SUCCESS (status)) 
			g_mlx4_dbg_level = g.bwsd.DebugPrintLevel = value;
		
		status = WdfRegistryQueryULong(hKey, &debugFlags, &value);
		if (NT_SUCCESS (status)) 
			g_mlx4_dbg_flags = g.bwsd.DebugPrintFlags = value;

		status = WdfRegistryQueryULong(hKey, &numQp, &value);
		if (NT_SUCCESS (status)) 
			g.log_num_qp = value;

		status = WdfRegistryQueryULong(hKey, &numRdmaRc, &value);
		if (NT_SUCCESS (status)) 
			g.log_rdmarc_per_qp = value;

		status = WdfRegistryQueryULong(hKey, &numSrq, &value);
		if (NT_SUCCESS (status)) 
			g.log_num_srq = value;

		status = WdfRegistryQueryULong(hKey, &numCq, &value);
		if (NT_SUCCESS (status)) 
			g.log_num_cq = value;

		status = WdfRegistryQueryULong(hKey, &numMcg, &value);
		if (NT_SUCCESS (status)) 
			g.log_num_mcg = value;

		status = WdfRegistryQueryULong(hKey, &numMpt, &value);
		if (NT_SUCCESS (status)) 
			g.log_num_mpt = value;
		
		status = WdfRegistryQueryULong(hKey, &numMtt, &value);
		if (NT_SUCCESS (status)) 
			g.log_num_mtt = value;

		status = WdfRegistryQueryULong(hKey, &numMac, &value);
		if (NT_SUCCESS (status)) 
			g.log_num_mac = value;
		else
			g.log_num_mac = 0;

		status = WdfRegistryQueryULong(hKey, &numVlan, &value);
		if (NT_SUCCESS (status)) 
			g.log_num_vlan= value;
		else
			g.log_num_vlan = 0;

		status = WdfRegistryQueryULong(hKey, &usePrio, &value);
		if (NT_SUCCESS (status)) 
			g.use_prio= value;
		else
			g.use_prio = 0;

		status = WdfRegistryQueryULong(hKey, &numFcExch, &value);
		if (NT_SUCCESS (status)) 
			g.num_fc_exch= value;
		else
			g.num_fc_exch = 0;

		status = WdfRegistryQueryULong(hKey, &enableQoS, &value);
		if (NT_SUCCESS (status)) 
			g.enable_qos = value;
		else
			g.enable_qos = 0;

		status = WdfRegistryQueryULong(hKey, &BlockMcastLB, &value);
		if (NT_SUCCESS (status)) 
			g.mlx4_blck_lb = value;
		else
			g.mlx4_blck_lb = 1;

		status = WdfRegistryQueryULong(hKey, &LogMttsPerSeg, &value);
		if (NT_SUCCESS (status)) 
			g.log_mtts_per_seg = value;
		else
			g.log_mtts_per_seg = ilog2(MLX4_MTT_ENTRY_PER_SEG);

		status = WdfRegistryQueryULong(hKey, &SrIov, &value);
		if (NT_SUCCESS (status)) 
			g.sr_iov = value;
		else
			g.sr_iov = 0;

		status = WdfRegistryQueryULong(hKey, &ProbeVf, &value);
		if (NT_SUCCESS (status)) 
			g.probe_vf = value;
		else
			g.probe_vf = 0;

		status = WdfRegistryQueryULong(hKey, &Set4kMtu, &value);
		if (NT_SUCCESS (status)) 
			g.set_4k_mtu = value;
		else
			g.set_4k_mtu = 0;

		status = WdfRegistryQueryULong(hKey, &InterruptFromFirstPacket, &value);
		if (NT_SUCCESS (status)) 
			g.interrupt_from_first = value;
		else
			g.interrupt_from_first = 1;

		status = WdfRegistryQueryULong(hKey, &ProcessorAffinity, &value);
		if (NT_SUCCESS (status)) 
			g.affinity = value;
		else
			g.affinity = 0;
		
		status = WdfRegistryQueryULong(hKey, &StatFlags, &value);
		if (NT_SUCCESS (status)) 
			g.stat_flags = value;
		else
			g.stat_flags = 0;
		
		status = WdfRegistryQueryULong(hKey, &ModeFlags, &value);
		if (NT_SUCCESS (status)) 
			g.mode_flags = value;
		else
			g.mode_flags = 0;
		
		status = WdfRegistryQueryULong(hKey, &SlaveNum, &value);
		if (NT_SUCCESS (status)) 
			g.slave_num = value;
		else
			g.slave_num = 0;

		status = WdfRegistryQueryULong(hKey, &EnableFcPreT11, &value);
		if (NT_SUCCESS (status)) 
			g.mlx4_pre_t11_mode = value;
		else
			g.mlx4_pre_t11_mode = 0;

		MLX4_PRINT(TRACE_LEVEL_WARNING, MLX4_DBG_DRV,
			("Globals: slave_num %d, mode_flags %x\n", g.slave_num, g.mode_flags));

		WdfRegistryClose(hKey);
		status = STATUS_SUCCESS;
	}

	return status;
}

NTSTATUS
DriverEntry(
	IN PDRIVER_OBJECT DriverObject,
	IN PUNICODE_STRING RegistryPath
	)
/*++
Routine Description:

	Initialize the call backs structure of Driver Framework.

Arguments:

	DriverObject - pointer to the driver object

	RegistryPath - pointer to a unicode string representing the path,
				to driver-specific key in the registry.

Return Value:

  NT Status Code

--*/
{
	int err;
	WDF_DRIVER_CONFIG	config;
	NTSTATUS			status;
	WDFDRIVER hDriver;

#if defined(EVENT_TRACING)
	WPP_INIT_TRACING(DriverObject, RegistryPath);
#endif


	RtlZeroMemory( &g, sizeof(g) );
	g_mlx4_dbg_level = g.bwsd.DebugPrintLevel = TRACE_LEVEL_VERBOSE;
	g_mlx4_dbg_flags = g.bwsd.DebugPrintFlags = 0xffff;
	g_time_increment = KeQueryTimeIncrement();
	KeInitializeEvent( &g.start_event, SynchronizationEvent , TRUE);
	//BUGBUG: SRIOV is not supported
	KeInitializeEvent( &g.surprise_removal_event, SynchronizationEvent , TRUE);

	MLX4_ENTER(MLX4_DBG_DRV);
	MLX4_PRINT(TRACE_LEVEL_INFORMATION, MLX4_DBG_DRV, 
		("Built %s %s, Version %s, RelDate %s\n", 
		__DATE__, __TIME__, DRV_VERSION, DRV_RELDATE));

	//
	// Initiialize driver config to control the attributes that
	// are global to the driver. Note that framework by default
	// provides a driver unload routine. If you create any resources
	// in the DriverEntry and want to be cleaned in driver unload,
	// you can override that by specifing one in the Config structure.
	//

	WDF_DRIVER_CONFIG_INIT(
		&config, EvtDriverDeviceAdd );
	config.EvtDriverUnload = EvtDriverUnload;

	//
	// Create a framework driver object to represent our driver.
	//
	status = WdfDriverCreate(DriverObject,
		RegistryPath, WDF_NO_OBJECT_ATTRIBUTES,
		&config, &hDriver);

	if (!NT_SUCCESS(status)) {
		MLX4_PRINT(TRACE_LEVEL_VERBOSE, MLX4_DBG_DRV, ("WdfDriverCreate failed with status 0x%x\n", status));
		goto end;
	}

	// get Registry parameters
	status = __read_registry(&hDriver);
	if ( !NT_SUCCESS( status ) )
		goto end;

	// global initializations
	err = mlx4_net_init();
	if (err) {
		status = errno_to_ntstatus(err);
		goto end;
	}
	err = core_init();
	if (err) {
		status = errno_to_ntstatus(err);
		goto end;
	}
	err = mlx4_ib_init();
	if (err) {
		status = errno_to_ntstatus(err);
		goto end;
	}

	// statistics
	RtlZeroMemory( &g_stat, sizeof(g_stat) );
	g_stat.drv.p_globals = &g;
	g_stat.drv.h_wdf_driver = hDriver;
	
	// we don't matter the failure in the work with Registry
	status = STATUS_SUCCESS;
	
end:
	MLX4_EXIT( MLX4_DBG_DRV );
	return status;

}




