/*
 * Copyright (c) 2005 SilverStorm Technologies.  All rights reserved.
 * Copyright (c) 1996-2003 Intel Corporation. All rights reserved. 
 *
 * This software is available to you under the OpenIB.org BSD license
 * below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * $Id: al.c 1611 2006-08-20 14:48:55Z sleybo $
 */

#include "precomp.h"
#include <initguid.h>
#include <wdmguid.h>
#include <rdma\verbs.h>

#if defined(EVENT_TRACING)
#ifdef offsetof
#undef offsetof
#endif
#include "drv.tmh"
#endif 

GLOBALS g;

/*
 * UVP name does not include file extension.  For debug builds, UAL
 * will append "d.dll".  For release builds, UAL will append ".dll"
 */
char			mlnx_uvp_lib_name[MAX_LIB_NAME] = {"mlx4u"};

static int __get_dev_info(PFDO_DEVICE_DATA p_fdo, __be64 *node_guid, u32 *hw_id)
{
	struct ib_device_attr device_attr;
	struct ib_device *p_ibdev = p_fdo->bus_ib_ifc.p_ibdev;
	int err;

	HCA_ENTER( HCA_DBG_PNP );

	err = (p_ibdev->query_device)( p_ibdev, &device_attr );
	if (err)
		return err;

	*node_guid = p_ibdev->node_guid;
	*hw_id = device_attr.hw_ver;
	HCA_EXIT( HCA_DBG_PNP );
	return 0;
}

#ifndef USE_WDM_FRAMEWORK

static ci_interface_t*
__alloc_hca_ifc(
	IN				PFDO_DEVICE_DATA const		p_fdo )
{
	ci_interface_t	*pIfc;

	HCA_ENTER( HCA_DBG_PNP );

	pIfc =
		(ci_interface_t*)ExAllocatePoolWithTag( NonPagedPool, sizeof(ci_interface_t), MT_TAG_KERNEL );
	if( !pIfc )
	{
		HCA_PRINT( TRACE_LEVEL_ERROR,HCA_DBG_PNP, 
			("Failed to allocate ci_interface_t (%d bytes).\n",
			sizeof(ci_interface_t)));
		return NULL;
	}

	setup_ci_interface( p_fdo->hca.guid, pIfc );

	pIfc->p_hca_obj = &p_fdo->hca;
	pIfc->vend_id = (uint32_t)p_fdo->bus_ib_ifc.pdev->ven_id;
	pIfc->dev_id = (uint16_t)p_fdo->bus_ib_ifc.pdev->dev_id;
	pIfc->dev_revision = (uint16_t)p_fdo->hca.hw_ver;

	HCA_EXIT( HCA_DBG_PNP );
	return pIfc;
}

static void
__unmap_hca_memory(
	IN				PFDO_DEVICE_DATA const p_fdo )
{
	struct pci_dev *pdev;
	int				i;

	HCA_ENTER( HCA_DBG_PNP );

	if (p_fdo->bus_ib_ifc_taken) {
		pdev = p_fdo->bus_ib_ifc.pdev;

		for( i = 0; i < HCA_BAR_TYPE_MAX; i++ ) {
			if (pdev->bar[i].virt) {
				MmUnmapIoSpace( pdev->bar[i].virt, pdev->bar[i].size );
				cl_memclr( &pdev->bar[i], sizeof(hca_bar_t) );
			}
		}
	}

	HCA_EXIT( HCA_DBG_PNP );
}

/* release the resources, allocated in hca_start */
static void
__hca_release_resources(
	IN	WDFDEVICE  Device )
{
	PFDO_DEVICE_DATA p_fdo = FdoGetData(Device);

	HCA_ENTER( HCA_DBG_PNP );

	switch( p_fdo->state )
	{
	case HCA_STARTED:
		/* dequeue HCA  */
		mlnx_hca_remove( &p_fdo->hca );
	}
	
	__unmap_hca_memory( p_fdo );

	p_fdo->state = HCA_ADDED;

	HCA_EXIT( HCA_DBG_PNP );
}

static void
__ref_ifc(
	IN				DEVICE_OBJECT*				p_dev_obj )
{
	PFDO_DEVICE_DATA p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;

	HCA_ENTER( HCA_DBG_PNP );

	cl_atomic_inc( &p_fdo->n_hca_ifc_ref );
	ObReferenceObject( p_dev_obj );

	HCA_PRINT( TRACE_LEVEL_INFORMATION, HCA_DBG_PNP, 
		("MLX4_HCA: CA_guid %I64x, hca_ifc_ref %d\n",
		p_fdo->hca.guid, p_fdo->n_hca_ifc_ref) );

	HCA_EXIT( HCA_DBG_PNP );
}

static void
__deref_ifc(
	IN				DEVICE_OBJECT*				p_dev_obj )
{
	PFDO_DEVICE_DATA p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;

	HCA_ENTER( HCA_DBG_PNP );

	cl_atomic_dec( &p_fdo->n_hca_ifc_ref );
	ObDereferenceObject( p_dev_obj );

	HCA_PRINT( TRACE_LEVEL_INFORMATION, HCA_DBG_PNP, 
		("MLX4_HCA: CA_guid %I64x, hca_ifc_ref %d\n",
		p_fdo->hca.guid, p_fdo->n_hca_ifc_ref) );

	HCA_EXIT( HCA_DBG_PNP );
}



NTSTATUS
EvtDeviceD0Entry(
	IN WDFDEVICE  Device,
	IN WDF_POWER_DEVICE_STATE  PreviousState
	)
{
	NTSTATUS status;
	PFDO_DEVICE_DATA p_fdo = FdoGetData(Device);

	HCA_PRINT(TRACE_LEVEL_ERROR, HCA_DBG_PNP,("REF_CNT_DBG: EvtDeviceD0Entry\n"));

	HCA_ENTER( HCA_DBG_PNP );
	HCA_PRINT(TRACE_LEVEL_INFORMATION, HCA_DBG_PNP, ("EvtDeviceD0Entry: PreviousState 0x%x\n", PreviousState));
	// start card (needed after Hibernation or standby)
	if (PreviousState > WdfPowerDeviceD0) {
		/* get MLX4_BUS IB interface */
		if ( !p_fdo->bus_ib_ifc_taken ) {
			status = WdfFdoQueryForInterface( Device, &MLX4_BUS_IB_INTERFACE_GUID,
				(PINTERFACE)&p_fdo->bus_ib_ifc, sizeof(MLX4_BUS_IB_INTERFACE), MLX4_BUS_IB_INTERFACE_VERSION, NULL );
			if( !NT_SUCCESS( status ) ) {
				HCA_PRINT(TRACE_LEVEL_ERROR, HCA_DBG_PNP, ("Getting MLX4 BUS interface failed: status=0x%x\n", status));
				return status;
			}
			p_fdo->bus_ib_ifc_taken = TRUE;
			p_fdo->bus_ib_ifc.p_ibdev->x.p_fdo = p_fdo;
		}
	}
	HCA_EXIT( HCA_DBG_PNP );
	return STATUS_SUCCESS;
}

NTSTATUS
EvtDeviceD0Exit(
	IN WDFDEVICE  Device,
	IN WDF_POWER_DEVICE_STATE  TargetState
	)
{
	NTSTATUS status;
	PFDO_DEVICE_DATA p_fdo = FdoGetData(Device);

	HCA_PRINT(TRACE_LEVEL_ERROR, HCA_DBG_PNP,("REF_CNT_DBG: EvtDeviceD0Exit\n"));

	HCA_ENTER( HCA_DBG_PNP );

	HCA_PRINT(TRACE_LEVEL_INFORMATION, HCA_DBG_PNP, ("EvtDeviceD0Exit: TargetState 0x%x\n", TargetState));

	switch (TargetState) {
	case WdfPowerDeviceD1:	/* hopefully, it is STANDBY state */
	case WdfPowerDeviceD2:	/* hopefully, it is STANDBY state */
	case WdfPowerDeviceD3:	/* hopefully, it is STANDBY state */
	case WdfPowerDevicePrepareForHibernation:
		if (atomic_read(&p_fdo->usecnt)) {
			status = STATUS_UNSUCCESSFUL;
			break;
		}
		/* Fall through. */
	default:
		status = STATUS_SUCCESS;
		break;
	}

	if (TargetState > WdfPowerDeviceD0) {
		if(p_fdo->bus_ib_ifc_taken) {
			PINTERFACE p_ifc = (PINTERFACE)&p_fdo->bus_ib_ifc;
			p_ifc->InterfaceDereference( p_ifc->Context );
			p_fdo->bus_ib_ifc_taken = FALSE;
		}
	}

	HCA_EXIT( HCA_DBG_PNP );
	return status;
}

NTSTATUS
EvtDevicePrepareHardware(
	IN WDFDEVICE  Device,
	IN WDFCMRESLIST  ResourcesRaw,
	IN WDFCMRESLIST  ResourcesTranslated
	)
{
	int err;
	NTSTATUS status;
	PFDO_DEVICE_DATA p_fdo  = FdoGetData(Device);
	BUS_INTERFACE_STANDARD	bus_pci_ifc;
	ci_interface_t *p_hca_ifc;
	RDMA_INTERFACE_VERBS rdma_ifc, *p_ifc = &rdma_ifc;
	WDF_QUERY_INTERFACE_CONFIG qiConfig;
		
	UNUSED_PARAM(ResourcesRaw);
	UNUSED_PARAM(ResourcesTranslated);

	HCA_ENTER( HCA_DBG_PNP );
	
	HCA_PRINT(TRACE_LEVEL_INFORMATION, HCA_DBG_PNP, ("EvtPrepareHardware: \n"));

	/* get PCI BUS interface */
	status = WdfFdoQueryForInterface( Device, &GUID_BUS_INTERFACE_STANDARD,
		(PINTERFACE)&bus_pci_ifc, sizeof(BUS_INTERFACE_STANDARD), 1, NULL );
	if( !NT_SUCCESS( status ) ) {
		HCA_PRINT(TRACE_LEVEL_ERROR, HCA_DBG_PNP, ("Getting PCI BUS interface failed: status=0x%x\n", status));
		return status;
	}
	RtlCopyMemory( &p_fdo->bus_pci_ifc, &bus_pci_ifc, sizeof(BUS_INTERFACE_STANDARD) );
	p_fdo->bus_pci_ifc_taken = TRUE;
	
	/* get MLX4_BUS IB interface */
	if ( !p_fdo->bus_ib_ifc_taken ) {
		status = WdfFdoQueryForInterface( Device, &MLX4_BUS_IB_INTERFACE_GUID,
			(PINTERFACE)&p_fdo->bus_ib_ifc, sizeof(MLX4_BUS_IB_INTERFACE), MLX4_BUS_IB_INTERFACE_VERSION, NULL );
		if( !NT_SUCCESS( status ) ) {
			HCA_PRINT(TRACE_LEVEL_ERROR, HCA_DBG_PNP, ("Getting MLX4 BUS interface failed: status=0x%x\n", status));
			return status;
		}
		p_fdo->bus_ib_ifc_taken = TRUE;
		p_fdo->bus_ib_ifc.p_ibdev->x.p_fdo = p_fdo;
	}

	InitializeListHead(&p_fdo->hca.event_list);
	KeInitializeSpinLock(&p_fdo->hca.event_list_lock);

	/* get node GUID */
	err = __get_dev_info( p_fdo, &p_fdo->hca.guid, &p_fdo->hca.hw_ver );
	if (err) {

		HCA_PRINT(TRACE_LEVEL_ERROR   ,HCA_DBG_LOW   ,
			("can't get guid - ib_query_device() failed (%08X)\n", err ));
		//TODO: no cleanup on error
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	/* queue HCA  */
	mlnx_hca_insert( &p_fdo->hca );

	/* Allocate and populate our HCA interface structure. */
	p_hca_ifc = __alloc_hca_ifc( p_fdo );
	if( !p_hca_ifc ) {
		HCA_PRINT( TRACE_LEVEL_ERROR  ,HCA_DBG_PNP  ,("__alloc_hca_ifc failed.\n"));
		return STATUS_NO_MEMORY;
	}

	/* fill interface fields */
	p_ifc->InterfaceHeader.Size = sizeof(RDMA_INTERFACE_VERBS);
	p_ifc->InterfaceHeader.Version = VerbsVersion(VERBS_MAJOR_VER, VERBS_MINOR_VER);
	p_ifc->InterfaceHeader.Context = p_fdo->p_dev_obj;
	p_ifc->InterfaceHeader.InterfaceReference = __ref_ifc;
	p_ifc->InterfaceHeader.InterfaceDereference = __deref_ifc;
	p_ifc->Verbs = *p_hca_ifc;
	ExFreePool( p_hca_ifc );

	/* create an upper interface */
	WDF_QUERY_INTERFACE_CONFIG_INIT( &qiConfig, (PINTERFACE)p_ifc,
		&GUID_RDMA_INTERFACE_VERBS, NULL);

	status = WdfDeviceAddQueryInterface( Device, &qiConfig );
	if (!NT_SUCCESS(status)) {
		HCA_PRINT( TRACE_LEVEL_ERROR  ,HCA_DBG_PNP  ,("WdfDeviceAddQueryInterface failed %#x\n", status));
		return status;
	}

	/*
	 * Change the state since the PnP callback can happen
	 * before the callback returns.
	 */
	p_fdo->state = HCA_STARTED;
	
	HCA_EXIT( HCA_DBG_PNP );
	return STATUS_SUCCESS;
}


NTSTATUS
EvtDeviceReleaseHardware(
	IN WDFDEVICE  Device,
	IN WDFCMRESLIST  ResourcesTranslated
	)
{
	PFDO_DEVICE_DATA p_fdo  = FdoGetData(Device);

	UNUSED_PARAM(ResourcesTranslated);

	HCA_ENTER( HCA_DBG_PNP );

	HCA_PRINT(TRACE_LEVEL_INFORMATION, HCA_DBG_PNP, ("EvtDeviceReleaseHardware: FdoData=0x%p\n", p_fdo));

	// release IBBUS resources
	__hca_release_resources(Device);

	// release MLX4_BUS resources
	if(p_fdo->bus_ib_ifc_taken) {
		PINTERFACE p_ifc = (PINTERFACE)&p_fdo->bus_ib_ifc;
		p_ifc->InterfaceDereference( p_ifc->Context );
		p_fdo->bus_ib_ifc_taken = FALSE;
	}

	// release PCI BUS resources
	if(p_fdo->bus_pci_ifc_taken) {
		PINTERFACE p_ifc = (PINTERFACE)&p_fdo->bus_pci_ifc;
		p_ifc->InterfaceDereference( p_ifc->Context );
		p_fdo->bus_pci_ifc_taken = FALSE;
	}

	HCA_EXIT( HCA_DBG_PNP );
	return STATUS_SUCCESS;
}

NTSTATUS
EvtDeviceQueryRemove(
	IN WDFDEVICE  Device
	)
{
	PFDO_DEVICE_DATA p_fdo = FdoGetData(Device);
	HCA_ENTER( HCA_DBG_PNP );
	if (atomic_read(&p_fdo->usecnt)) {
		cl_dbg_out( "MLX4_HCA: Can't get unloaded. %d applications are still in work\n", p_fdo->usecnt);
		return STATUS_UNSUCCESSFUL;
	}
	HCA_EXIT( HCA_DBG_PNP );
	return STATUS_SUCCESS;
}


NTSTATUS
EvtDriverDeviceAdd(
	IN WDFDRIVER        Driver,
	IN PWDFDEVICE_INIT  DeviceInit
	)
/*++
Routine Description:

	EvtDriverDeviceAdd is called by the framework in response to AddDevice
	call from the PnP manager. We create and initialize a device object to
	represent a new instance of mxe bus.

Arguments:

	Driver - Handle to a framework driver object created in DriverEntry

	DeviceInit - Pointer to a framework-allocated WDFDEVICE_INIT structure.

Return Value:

	NTSTATUS

--*/
{
	WDF_OBJECT_ATTRIBUTES			attributes;
	NTSTATUS						status;
	WDFDEVICE						device;
	PFDO_DEVICE_DATA				p_fdo;
	WDF_PNPPOWER_EVENT_CALLBACKS	Callbacks;

	UNREFERENCED_PARAMETER(Driver);

	PAGED_CODE ();
	
	HCA_PRINT(TRACE_LEVEL_INFORMATION, HCA_DBG_PNP, ("EvtDriverDeviceAdd: 0x%p\n", Driver));
	//
	// register PnP & Power stuff
	//
	WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&Callbacks);
	Callbacks.EvtDevicePrepareHardware = EvtDevicePrepareHardware;
	Callbacks.EvtDeviceReleaseHardware = EvtDeviceReleaseHardware;
	Callbacks.EvtDeviceQueryRemove  = EvtDeviceQueryRemove;
	Callbacks.EvtDeviceD0Entry = EvtDeviceD0Entry;
	Callbacks.EvtDeviceD0Exit = EvtDeviceD0Exit;

	WdfDeviceInitSetPnpPowerEventCallbacks( DeviceInit, &Callbacks );
	
	//
	// Initialize all the properties specific to the device.
	// Framework has default values for the one that are not
	// set explicitly here. So please read the doc and make sure
	// you are okay with the defaults.
	//
	WdfDeviceInitSetDeviceType(DeviceInit, FILE_DEVICE_INFINIBAND);
	WdfDeviceInitSetExclusive(DeviceInit, FALSE);

	//
	// Initialize attributes structure to specify size and accessor function
	// for storing device context.
	//
	WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attributes, FDO_DEVICE_DATA);

	//
	// Create a framework device object. In response to this call, framework
	// creates a WDM deviceobject.
	//
	status = WdfDeviceCreate(&DeviceInit, &attributes, &device);
	if (!NT_SUCCESS(status)) {
		HCA_PRINT(TRACE_LEVEL_VERBOSE, HCA_DBG_PNP, ("EvtDriverDeviceAdd: WdfDeviceCreate failed with 0x%x\n", status));
		goto end;
	}

	//
	// Init device context.
	//
	p_fdo = FdoGetData(device);
	RtlZeroMemory(p_fdo, sizeof(FDO_DEVICE_DATA));
	p_fdo->FdoDevice = device;
	p_fdo->p_dev_obj = WdfDeviceWdmGetDeviceObject( device );
	spin_lock_init( &p_fdo->uctx_lock );
	cl_qlist_init( &p_fdo->uctx_list );
	atomic_set(&p_fdo->usecnt, 0);
	p_fdo->state = HCA_ADDED;

	//
	// WMI
	//
	status = WmiRegistration(device);
	if (!NT_SUCCESS(status)) {
		return status;
	}

	status = STATUS_SUCCESS;

end:	
	HCA_EXIT( HCA_DBG_PNP );
	return status;
}


void
EvtDriverUnload(
	IN		WDFDRIVER  Driver
	)
{
	HCA_ENTER( HCA_DBG_PNP );

	UNUSED_PARAM( Driver );

	HCA_EXIT( HCA_DBG_PNP );
#if defined(EVENT_TRACING)
	WPP_CLEANUP(WdfDriverWdmGetDriverObject(Driver));
#endif
}

NTSTATUS
DriverEntry(
	IN PDRIVER_OBJECT DriverObject,
	IN PUNICODE_STRING RegistryPath
	)
/*++
Routine Description:

	Initialize the call backs structure of Driver Framework.

Arguments:

	DriverObject - pointer to the driver object

	RegistryPath - pointer to a unicode string representing the path,
					to driver-specific key in the registry.

Return Value:

	NT Status Code

--*/
{
	WDF_DRIVER_CONFIG   config;
	NTSTATUS            status;
	WDFDRIVER hDriver;

#if defined(EVENT_TRACING)
	WPP_INIT_TRACING(DriverObject, RegistryPath);
#endif

	// global initializations
	g.DebugPrintLevel = TRACE_LEVEL_VERBOSE;
	g.DebugPrintFlags = 0xffff;
	HCA_ENTER( HCA_DBG_PNP );

	status = mlnx_hcas_init();
	if( status  != STATUS_SUCCESS ) {
		HCA_PRINT( TRACE_LEVEL_ERROR ,HCA_DBG_PNP ,
			("mlnx_hcas_init returned %#x.\n", status));
		goto end;
	}

	//
	// Initiialize driver config to control the attributes that
	// are global to the driver. Note that framework by default
	// provides a driver unload routine. If you create any resources
	// in the DriverEntry and want to be cleaned in driver unload,
	// you can override that by specifing one in the Config structure.
	//

	WDF_DRIVER_CONFIG_INIT(
		&config, EvtDriverDeviceAdd );
	config.EvtDriverUnload = EvtDriverUnload;

	//
	// Create a framework driver object to represent our driver.
	//
	status = WdfDriverCreate(DriverObject,
		RegistryPath, WDF_NO_OBJECT_ATTRIBUTES,
		&config, &hDriver);

	if (!NT_SUCCESS(status)) {
		HCA_PRINT(TRACE_LEVEL_VERBOSE, HCA_DBG_PNP, ("WdfDriverCreate failed with status 0x%x\n", status));
		goto end;
	}

	//
	// read registry parameters
	//
	{
		DECLARE_CONST_UNICODE_STRING(valueName0, L"DebugLevel");
		DECLARE_CONST_UNICODE_STRING(valueName1, L"DebugFlags");
		ULONG value;
		WDFKEY hKey = NULL;

		status = WdfDriverOpenParametersRegistryKey( hDriver,
			STANDARD_RIGHTS_ALL, WDF_NO_OBJECT_ATTRIBUTES, &hKey );

		if (NT_SUCCESS (status)) {

			status = WdfRegistryQueryULong(hKey, &valueName0, &value);
			if (NT_SUCCESS (status)) g.DebugPrintLevel = value;
			
			status = WdfRegistryQueryULong(hKey, &valueName1, &value);
			if (NT_SUCCESS (status)) g.DebugPrintFlags = value;
			
			WdfRegistryClose(hKey);
		}

		// we don't matter the failure in the work with Registry
		status = STATUS_SUCCESS;
	}

end:
	HCA_PRINT( TRACE_LEVEL_INFORMATION,HCA_DBG_PNP ,
		("exit status %#x.\n", status));
	HCA_EXIT( HCA_DBG_PNP );
	return status;

}

#else


UNICODE_STRING				g_param_path;
static cl_vfptr_pnp_po_t	vfptrHcaPnp;

NTSTATUS
DriverEntry(
	IN				PDRIVER_OBJECT				p_driver_obj,
	IN				PUNICODE_STRING				p_registry_path );

static NTSTATUS
__read_registry(
	IN				UNICODE_STRING* const		p_Param_Path );

static void
hca_drv_unload(
	IN				PDRIVER_OBJECT				p_driver_obj );

static NTSTATUS
hca_sysctl(
	IN				PDEVICE_OBJECT				p_dev_obj,
	IN				PIRP						p_irp );

NTSTATUS
hca_add_device(
	IN				PDRIVER_OBJECT				pDriverObj,
	IN				PDEVICE_OBJECT				pPdo );

static NTSTATUS
hca_start(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_query_stop(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_stop(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_cancel_stop(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_query_remove(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static void
hca_release_resources(
	IN				DEVICE_OBJECT* const		p_dev_obj );

static NTSTATUS
hca_cancel_remove(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_surprise_remove(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_query_capabilities(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_query_interface(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_query_pnp_state(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_query_bus_relations(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_query_removal_relations(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_query_power(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp,
		OUT			cl_irp_action_t* const		p_action );

static NTSTATUS
hca_set_power(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action );

static ci_interface_t*
__alloc_hca_ifc(
	IN				FDO_DEVICE_DATA* const		p_fdo );

static NTSTATUS
__get_ci_interface(
	IN				DEVICE_OBJECT* const		p_dev_obj );




NTSTATUS
hca_add_device(
	IN				PDRIVER_OBJECT				pDriverObj,
	IN				PDEVICE_OBJECT				pPdo )
{
	NTSTATUS			status;
	DEVICE_OBJECT		*p_dev_obj, *pNextDevObj;
	PFDO_DEVICE_DATA	p_fdo;

	HCA_ENTER(HCA_DBG_PNP);

	/*
	 * Create the device so that we have a device extension to store stuff in.
	 */
	status = IoCreateDevice( pDriverObj, sizeof(FDO_DEVICE_DATA),
		NULL, FILE_DEVICE_INFINIBAND, FILE_DEVICE_SECURE_OPEN,
		FALSE, &p_dev_obj );
	if( !NT_SUCCESS( status ) )
	{
		HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PNP, 
			("IoCreateDevice returned 0x%08X.\n", status));
		return status;
	}

	p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;
	cl_memclr( p_fdo, sizeof(FDO_DEVICE_DATA) );
	cl_spinlock_init( &p_fdo->uctx_lock );
	cl_qlist_init( &p_fdo->uctx_list );
	atomic_set(&p_fdo->usecnt, 0);

	/* Attach to the device stack. */
	pNextDevObj = IoAttachDeviceToDeviceStack( p_dev_obj, pPdo );
	if( !pNextDevObj )
	{
		//cl_event_destroy( &p_fdo->mutex );
		IoDeleteDevice( p_dev_obj );
		HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PNP, 
			("IoAttachDeviceToDeviceStack failed.\n"));
		return STATUS_NO_SUCH_DEVICE;
	}

	/* Inititalize the complib extension. */
	cl_init_pnp_po_ext( p_dev_obj, pNextDevObj, pPdo, CL_DBG_PNP | CL_DBG_ERROR,
		&vfptrHcaPnp, NULL );

	p_fdo->state = HCA_ADDED;

	HCA_EXIT(HCA_DBG_PNP);
	return status;
}

static void
__put_ifc(
		IN	PINTERFACE		p_ifc )
{
	HCA_ENTER( HCA_DBG_PNP );
	p_ifc->InterfaceDereference( p_ifc->Context );
	HCA_EXIT( HCA_DBG_PNP );
}

/* Forwards the request to the HCA's PDO. */
static NTSTATUS
__get_ifc(
	IN				DEVICE_OBJECT* const		pDevObj,
	IN		const	GUID* const					pGuid,
	IN				USHORT						size,
	IN				USHORT						Version,
	IN OUT			PVOID						InterfaceSpecificData,
		OUT			PINTERFACE					pBusIfc )
{
	NTSTATUS			status;
	IRP					*pIrp;
	IO_STATUS_BLOCK		ioStatus;
	IO_STACK_LOCATION	*pIoStack;
	DEVICE_OBJECT		*pDev;
	KEVENT				event;

	HCA_ENTER( HCA_DBG_PNP );

	CL_ASSERT( KeGetCurrentIrql() < DISPATCH_LEVEL );

	pDev = IoGetAttachedDeviceReference( pDevObj );

	KeInitializeEvent( &event, NotificationEvent, FALSE );

	/* Build the IRP for the HCA. */
	pIrp = IoBuildSynchronousFsdRequest( IRP_MJ_PNP, pDev,
		NULL, 0, NULL, &event, &ioStatus );
	if( !pIrp )
	{
		ObDereferenceObject( pDev );
		HCA_PRINT( TRACE_LEVEL_ERROR,HCA_DBG_PNP, 
			("IoBuildSynchronousFsdRequest failed.\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	/* Copy the request query parameters. */
	pIoStack = IoGetNextIrpStackLocation( pIrp );
	pIoStack->MinorFunction = IRP_MN_QUERY_INTERFACE;
	pIoStack->Parameters.QueryInterface.Size = size;
	pIoStack->Parameters.QueryInterface.Version = Version;
	pIoStack->Parameters.QueryInterface.InterfaceType = pGuid;
	pIoStack->Parameters.QueryInterface.Interface = (INTERFACE*)pBusIfc;
	pIoStack->Parameters.QueryInterface.InterfaceSpecificData = InterfaceSpecificData;

	pIrp->IoStatus.Status = STATUS_NOT_SUPPORTED;

	/* Send the IRP. */
	status = IoCallDriver( pDev, pIrp );
	if( status == STATUS_PENDING )
	{
		KeWaitForSingleObject( &event, Executive, KernelMode,
			FALSE, NULL );

		status = ioStatus.Status;
	}
	ObDereferenceObject( pDev );

	HCA_EXIT( HCA_DBG_PNP );
	return status;
}

static ci_interface_t*
__alloc_hca_ifc(
	IN				PFDO_DEVICE_DATA const		p_fdo )
{
	ci_interface_t	*pIfc;

	HCA_ENTER( HCA_DBG_PNP );

	pIfc = (ci_interface_t*)ExAllocatePoolWithTag( NonPagedPool,
		sizeof(ci_interface_t), MT_TAG_KERNEL );
	if( !pIfc )
	{
		HCA_PRINT( TRACE_LEVEL_ERROR,HCA_DBG_PNP, 
			("Failed to allocate ci_interface_t (%d bytes).\n",
			sizeof(ci_interface_t)));
		return NULL;
	}

	setup_ci_interface( p_fdo->hca.guid, pIfc );

	pIfc->p_hca_obj = &p_fdo->hca;
	pIfc->vend_id = (uint32_t)p_fdo->bus_ib_ifc.pdev->ven_id;
	pIfc->dev_id = (uint16_t)p_fdo->bus_ib_ifc.pdev->dev_id;
	pIfc->dev_revision = (uint16_t)p_fdo->hca.hw_ver;

	HCA_EXIT( HCA_DBG_PNP );
	return pIfc;
}

static void
__unmap_hca_memory(
	IN				PFDO_DEVICE_DATA const p_fdo )
{
	struct pci_dev *pdev = p_fdo->bus_ib_ifc.pdev;
	int				i;

	HCA_ENTER( HCA_DBG_PNP );

	if ( pdev )
		for( i = 0; i < HCA_BAR_TYPE_MAX; i++ ) {
			if (pdev->bar[i].virt) {
				MmUnmapIoSpace( pdev->bar[i].virt, pdev->bar[i].size );
				cl_memclr( &pdev->bar[i], sizeof(hca_bar_t) );
			}
		}

	HCA_EXIT( HCA_DBG_PNP );
}


/* release the resources, allocated in hca_start */
static void
__hca_release_resources(
	IN				DEVICE_OBJECT* const		p_dev_obj )
{
	PFDO_DEVICE_DATA		p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;

	HCA_ENTER( HCA_DBG_PNP );

	switch( p_fdo->state )
	{
	case HCA_STARTED:
		/* dequeue HCA  */
		mlnx_hca_remove( &p_fdo->hca );
	}

	if (p_fdo->al_sym_name.Buffer) {
		ExFreePool( p_fdo->al_sym_name.Buffer );
		p_fdo->al_sym_name.Buffer = NULL;
	}
	
	if( p_fdo->pnp_target_entry )
	{
		ASSERT( p_fdo->pnp_ifc_entry );
		IoUnregisterPlugPlayNotification( p_fdo->pnp_target_entry );
		p_fdo->pnp_target_entry = NULL;
	}

	if( p_fdo->pnp_ifc_entry ) {
		IoUnregisterPlugPlayNotification( p_fdo->pnp_ifc_entry );
		p_fdo->pnp_ifc_entry = NULL;
	}

	if( p_fdo->p_al_file_obj ) {
		ObDereferenceObject( p_fdo->p_al_file_obj );
		p_fdo->p_al_file_obj = NULL;
	}

	// release MLX4_BUS resources
	if(p_fdo->bus_ib_ifc_taken) {
		p_fdo->bus_ib_ifc_taken = FALSE;
		__put_ifc( (PINTERFACE)&p_fdo->bus_ib_ifc );
	}

	// release PCI BUS resources
	if(p_fdo->bus_pci_ifc_taken) {
		p_fdo->bus_pci_ifc_taken = FALSE;
		__put_ifc( (PINTERFACE)&p_fdo->bus_pci_ifc );
	}

	__unmap_hca_memory( p_fdo );

	p_fdo->state = HCA_ADDED;

	HCA_EXIT( HCA_DBG_PNP );
}


static void
hca_release_resources(
	IN				DEVICE_OBJECT* const		p_dev_obj )
{
	PFDO_DEVICE_DATA		p_fdo;
	POWER_STATE		powerState;

	HCA_ENTER( HCA_DBG_PNP );

	p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;

	/* release all the resources, allocated in hca_start */
	__hca_release_resources(p_dev_obj);

	/* Notify the power manager that the device is powered down. */
	p_fdo->DevicePowerState = PowerDeviceD3;
	powerState.DeviceState = PowerDeviceD3;
	powerState = PoSetPowerState ( p_fdo->cl_ext.p_self_do, DevicePowerState, powerState );

	HCA_PRINT( TRACE_LEVEL_INFORMATION, HCA_DBG_PNP, 
		("PoSetPowerState: old state %d, new state to %d\n", 
		powerState.DeviceState, p_fdo->DevicePowerState ));

	/* Clear the PnP state in case we get restarted. */
	p_fdo->pnpState = 0;

	HCA_EXIT( HCA_DBG_PNP );
}

static NTSTATUS
hca_start(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action )
{
	int					err;
	NTSTATUS			status;
	PFDO_DEVICE_DATA	p_fdo;
	IO_STACK_LOCATION	*pIoStack;
	POWER_STATE			powerState;
	BUS_INTERFACE_STANDARD	bus_pci_ifc;

	HCA_ENTER( HCA_DBG_PNP );

	p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;

	/* Handled on the way up. */
	status = cl_do_sync_pnp( p_dev_obj, p_irp, p_action );
	if( !NT_SUCCESS( status ) )
	{
		HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PNP, 
			("Lower drivers failed IRP_MN_START_DEVICE (%#x).\n", status));
		goto end;
	}

	pIoStack = IoGetCurrentIrpStackLocation( p_irp );

	/* get PCI BUS interface */
	status = __get_ifc( p_dev_obj, &GUID_BUS_INTERFACE_STANDARD,
		sizeof(BUS_INTERFACE_STANDARD), 1, NULL, (PINTERFACE)&bus_pci_ifc);
	if( !NT_SUCCESS( status ) ) {
		HCA_PRINT(TRACE_LEVEL_ERROR, HCA_DBG_PNP, ("Getting PCI BUS interface failed: status=0x%x\n", status));
		goto end;
	}
	RtlCopyMemory( &p_fdo->bus_pci_ifc, &bus_pci_ifc, sizeof(BUS_INTERFACE_STANDARD) );
	p_fdo->bus_pci_ifc_taken = TRUE;
	
	/* get MLX4_BUS IB interface */
	status = __get_ifc( p_dev_obj, &MLX4_BUS_IB_INTERFACE_GUID,
		sizeof(MLX4_BUS_IB_INTERFACE), MLX4_BUS_IB_INTERFACE_VERSION, NULL, (PINTERFACE)&p_fdo->bus_ib_ifc);
	if( !NT_SUCCESS( status ) ) {
		HCA_PRINT(TRACE_LEVEL_ERROR, HCA_DBG_PNP, ("Getting MLX4 BUS interface failed: status=0x%x\n", status));
		goto end;
	}
	p_fdo->bus_ib_ifc_taken = TRUE;
	p_fdo->bus_ib_ifc.p_ibdev->x.p_fdo = p_fdo;
	
	InitializeListHead(&p_fdo->hca.event_list);
	KeInitializeSpinLock(&p_fdo->hca.event_list_lock);

	/* get node GUID */
	err = __get_dev_info( p_fdo, &p_fdo->hca.guid, &p_fdo->hca.hw_ver );
	if (err) {

		HCA_PRINT(TRACE_LEVEL_ERROR   ,HCA_DBG_LOW   ,
			("can't get guid - ib_query_device() failed (%08X)\n", err ));
		//TODO: no cleanup on error
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	/* queue HCA  */
	mlnx_hca_insert( &p_fdo->hca );

	/*
	 * Change the state since the PnP callback can happen
	 * before the callback returns.
	 */
	p_fdo->state = HCA_STARTED;
	
	/* We get started fully powered. */
	p_fdo->DevicePowerState = PowerDeviceD0;
	powerState.DeviceState = PowerDeviceD0;
	powerState = PoSetPowerState ( p_fdo->cl_ext.p_self_do, DevicePowerState, powerState );
	HCA_PRINT( TRACE_LEVEL_INFORMATION, HCA_DBG_PNP, 
		("PoSetPowerState: old state %d, new state to %d\n", 
		powerState.DeviceState, p_fdo->DevicePowerState ));

end:
	HCA_EXIT( HCA_DBG_PNP );
	return status;
}

static NTSTATUS
hca_query_stop(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action )
{
	/* All kernel clients will get notified through the device hierarchy. */

	/* TODO: set a flag to fail creation of any new IB resources. */
	return cl_irp_skip( p_dev_obj, p_irp, p_action );
}


static NTSTATUS
hca_stop(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action )
{
	/*
	 * Must disable everything.  Complib framework will
	 * call ReleaseResources handler.
	 */
	return cl_irp_skip( p_dev_obj, p_irp, p_action );
}


static NTSTATUS
hca_cancel_stop(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action )
{
	/* Handled on the way up. */
	return cl_do_sync_pnp( p_dev_obj, p_irp, p_action );
}


static NTSTATUS
hca_query_remove(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action )
{
	PFDO_DEVICE_DATA p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;
	if (atomic_read(&p_fdo->usecnt)) {
		cl_dbg_out( "MTHCA: Can't get unloaded. %d applications are still in work\n", p_fdo->usecnt);
		p_irp->IoStatus.Status = STATUS_UNSUCCESSFUL;
		return cl_irp_complete( p_dev_obj, p_irp, p_action );
	}
	/* TODO: set a flag to fail creation of any new IB resources. */
	return cl_irp_skip( p_dev_obj, p_irp, p_action );
}


static NTSTATUS
hca_cancel_remove(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action )
{
	/* Handled on the way up. */
	return cl_do_sync_pnp( p_dev_obj, p_irp, p_action );
}


static NTSTATUS
hca_surprise_remove(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action )
{
	/*
	 * TODO: Set state so that all further requests
	 * automatically succeed/fail as needed.
	 */
	return cl_irp_skip( p_dev_obj, p_irp, p_action );
}


static NTSTATUS
hca_query_capabilities(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action )
{
	NTSTATUS			status;
	PFDO_DEVICE_DATA	p_fdo;
	IO_STACK_LOCATION	*pIoStack;
	DEVICE_CAPABILITIES	*pCaps;

	HCA_ENTER( HCA_DBG_PNP );

	p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;

	/* Process on the way up. */
	status = cl_do_sync_pnp( p_dev_obj, p_irp, p_action );
	if( !NT_SUCCESS( status ) )
	{
		HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PNP, 
			("cl_do_sync_pnp returned %08X.\n", status));
		return status;
	}

	pIoStack = IoGetCurrentIrpStackLocation( p_irp );
	pCaps = pIoStack->Parameters.DeviceCapabilities.Capabilities;

	/*
	 * Store the device power mapping into our extension since we're
	 * the power policy owner.  The mapping is used when handling
	 * IRP_MN_SET_POWER IRPs.
	 */
	cl_memcpy(
		p_fdo->DevicePower, pCaps->DeviceState, sizeof(p_fdo->DevicePower) );

	if( pCaps->DeviceD1 )
	{
		HCA_PRINT( TRACE_LEVEL_WARNING ,HCA_DBG_PNP,
			("WARNING: Device reports support for DeviceD1 power state.\n"));
		pCaps->DeviceD1 = FALSE;
	}

	if( pCaps->DeviceD2 )
	{
		HCA_PRINT( TRACE_LEVEL_WARNING,HCA_DBG_PNP,
			("WARNING: Device reports support for DeviceD2 power state.\n"));
		pCaps->DeviceD2 = FALSE;
	}

	if( pCaps->SystemWake != PowerSystemUnspecified )
	{
		HCA_PRINT( TRACE_LEVEL_WARNING ,HCA_DBG_PNP,
			("WARNING: Device reports support for system wake.\n"));
		pCaps->SystemWake = PowerSystemUnspecified;
	}

	if( pCaps->DeviceWake != PowerDeviceUnspecified )
	{
		HCA_PRINT( TRACE_LEVEL_WARNING, HCA_DBG_PNP,
			("WARNING: Device reports support for device wake.\n"));
		pCaps->DeviceWake = PowerDeviceUnspecified;
	}

	HCA_EXIT( HCA_DBG_PNP );
	return status;
}

static void
__ref_ifc(
	IN				DEVICE_OBJECT*				p_dev_obj )
{
	PFDO_DEVICE_DATA p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;

	HCA_ENTER( HCA_DBG_PNP );

	cl_atomic_inc( &p_fdo->n_hca_ifc_ref );
	ObReferenceObject( p_dev_obj );

	HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PNP, 
		("MLX4_HCA: CA_guid %I64x, hca_ifc_ref %d\n",
		p_fdo->hca.guid, p_fdo->n_hca_ifc_ref) );

	HCA_EXIT( HCA_DBG_PNP );
}

static void
__deref_ifc(
	IN				DEVICE_OBJECT*				p_dev_obj )
{
	PFDO_DEVICE_DATA p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;

	HCA_ENTER( HCA_DBG_PNP );

	cl_atomic_dec( &p_fdo->n_hca_ifc_ref );
	ObDereferenceObject( p_dev_obj );

	HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PNP, 
		("MLX4_HCA: CA_guid %I64x, hca_ifc_ref %d\n",
		p_fdo->hca.guid, p_fdo->n_hca_ifc_ref) );

	HCA_EXIT( HCA_DBG_PNP );
}

static NTSTATUS
__query_ci_ifc(
	IN					DEVICE_OBJECT* const		p_dev_obj,
	IN					IO_STACK_LOCATION* const	p_io_stack )
{
	RDMA_INTERFACE_VERBS	*p_ifc;
	PFDO_DEVICE_DATA		p_fdo;
	ci_interface_t			*p_hca_ifc;
	NTSTATUS				status;
	UINT8					version;

	HCA_ENTER( HCA_DBG_PNP );

	version = VerbsVersionMajor(p_io_stack->Parameters.QueryInterface.Version);
	if(  version > VERBS_MAJOR_VER )
	{
		status = STATUS_NOT_SUPPORTED;
		goto exit;
	}

	if( p_io_stack->Parameters.QueryInterface.Size < sizeof(RDMA_INTERFACE_VERBS) )
	{
		status = STATUS_BUFFER_TOO_SMALL;
		goto exit;
	}

	p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;
	p_hca_ifc = __alloc_hca_ifc( p_fdo );
	if( !p_hca_ifc )
	{
		status = STATUS_NO_MEMORY;
		goto exit;
	}

	p_ifc = (RDMA_INTERFACE_VERBS *) p_io_stack->Parameters.QueryInterface.Interface;

	p_ifc->InterfaceHeader.Size = sizeof(RDMA_INTERFACE_VERBS);
	p_ifc->InterfaceHeader.Version = VerbsVersion(VERBS_MAJOR_VER, VERBS_MINOR_VER);
	p_ifc->InterfaceHeader.Context = p_dev_obj;
	p_ifc->InterfaceHeader.InterfaceReference = __ref_ifc;
	p_ifc->InterfaceHeader.InterfaceDereference = __deref_ifc;
	p_ifc->Verbs = *p_hca_ifc;

	/* take the reference before returning. */
	__ref_ifc( p_dev_obj );

	ExFreePool( p_hca_ifc );
	status = STATUS_SUCCESS;

exit:
	HCA_EXIT( HCA_DBG_PNP );
	return status;
}


static NTSTATUS
hca_query_interface(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action )
{
	NTSTATUS			status;
	IO_STACK_LOCATION	*p_io_stack;

	HCA_ENTER( HCA_DBG_PNP );

#pragma warning( push, 3 )
	PAGED_CODE();
#pragma warning( pop )

	p_io_stack = IoGetCurrentIrpStackLocation( p_irp );
	
	/* Compare requested GUID with our supported interface GUIDs. */
	if( IsEqualGUID( p_io_stack->Parameters.QueryInterface.InterfaceType,
		&GUID_RDMA_INTERFACE_VERBS ) )
	{
		status = __query_ci_ifc( p_dev_obj, p_io_stack );
		if( !NT_SUCCESS( status ) )
		{
			*p_action = IrpComplete;
		}
		else
		{
			*p_action = IrpSkip;
		}
	}
	else
	{
		status = p_irp->IoStatus.Status;
		*p_action = IrpSkip;
	}

	HCA_EXIT( HCA_DBG_PNP );
	return status;
}


static NTSTATUS
hca_query_pnp_state(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp, 
		OUT			cl_irp_action_t* const		p_action )
{
	PFDO_DEVICE_DATA		p_fdo;

	HCA_ENTER( HCA_DBG_PNP );

	p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;

	p_irp->IoStatus.Information |= p_fdo->pnpState;

	*p_action = IrpSkip;

	HCA_EXIT( HCA_DBG_PNP );
	return STATUS_SUCCESS;;
}

static NTSTATUS
hca_query_power(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp,
		OUT			cl_irp_action_t* const		p_action )
{
	NTSTATUS			status = STATUS_SUCCESS;
	IO_STACK_LOCATION	*pIoStack;

	HCA_ENTER(HCA_DBG_PO);

	UNUSED_PARAM( p_dev_obj );

	pIoStack = IoGetCurrentIrpStackLocation( p_irp );

	HCA_PRINT( TRACE_LEVEL_INFORMATION, HCA_DBG_PO, 
		("QUERY_POWER for FDO %p: type %s, state %d, action %d, IRQL %d, IRP %p\n",
		p_dev_obj, 
		(pIoStack->Parameters.Power.Type) ? "DevicePowerState" : "SystemPowerState",
		pIoStack->Parameters.Power.State.DeviceState, 
		pIoStack->Parameters.Power.ShutdownType, KeGetCurrentIrql(), p_irp ));

	switch( pIoStack->Parameters.Power.Type )
	{
	case SystemPowerState:
		/* Fail any requests to hibernate or sleep the system. */
		switch( pIoStack->Parameters.Power.State.SystemState )
		{
			case PowerSystemSleeping1:	// STANDBY support
			case PowerSystemSleeping2:	// STANDBY support
			case PowerSystemSleeping3:	// STANDBY support
			case PowerSystemHibernate:
			{
				PFDO_DEVICE_DATA p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;
				if (atomic_read(&p_fdo->usecnt)) 
					status = STATUS_UNSUCCESSFUL;
				break;
			}

			case PowerSystemWorking:
			case PowerSystemShutdown:
				break;

			default:
				status = STATUS_NOT_SUPPORTED;
		}
		break;

	case DevicePowerState:
		/* Fail any query for low power states. */
		switch( pIoStack->Parameters.Power.State.DeviceState )
		{
		case PowerDeviceD0:
		case PowerDeviceD3:
			/* We only support fully powered or off power states. */
			break;

		default:
			status = STATUS_NOT_SUPPORTED;
		}
		break;
	}

	if( status == STATUS_SUCCESS )
		*p_action = IrpSkip;
	else
		*p_action = IrpComplete;

	HCA_EXIT( HCA_DBG_PO );
	return status;
}


static void
__RequestPowerCompletion(
	IN				DEVICE_OBJECT				*p_dev_obj,
	IN				UCHAR						minorFunction,
	IN				POWER_STATE					powerState,
	IN				void						*context,
	IN				IO_STATUS_BLOCK				*pIoStatus )
{
	IRP					*p_irp;
	cl_pnp_po_ext_t		*p_fdo;

	HCA_ENTER( HCA_DBG_PO );

	UNUSED_PARAM( minorFunction );
	UNUSED_PARAM( powerState );

	p_irp = (IRP*)context;
	p_fdo = (cl_pnp_po_ext_t*)p_dev_obj->DeviceExtension;

	/* Propagate the device IRP status to the system IRP status. */
	p_irp->IoStatus.Status = pIoStatus->Status;

	/* Continue Power IRP processing. */
	PoStartNextPowerIrp( p_irp );
	IoCompleteRequest( p_irp, IO_NO_INCREMENT );
	IoReleaseRemoveLock( &p_fdo->remove_lock, p_irp );
	HCA_EXIT( HCA_DBG_PO );
}


/*NOTE: Completion routines must NEVER be pageable. */
static NTSTATUS
__SystemPowerCompletion(
	IN				DEVICE_OBJECT				*p_dev_obj,
	IN				IRP							*p_irp,
	IN				void						*context )
{
	NTSTATUS			status;
	POWER_STATE			state;
	PFDO_DEVICE_DATA	p_fdo;
	IO_STACK_LOCATION	*pIoStack;

	HCA_ENTER( HCA_DBG_PO );

	UNUSED_PARAM( context );

	p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;
	pIoStack = IoGetCurrentIrpStackLocation( p_irp );

	if( !NT_SUCCESS( p_irp->IoStatus.Status ) )
	{
		HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PO, 
			("IRP_MN_SET_POWER for system failed by lower driver with %08x.\n",
			p_irp->IoStatus.Status));
		status = STATUS_SUCCESS;
		PoStartNextPowerIrp( p_irp );
		goto release;
	}

	state.DeviceState = 
		p_fdo->DevicePower[pIoStack->Parameters.Power.State.SystemState];

	/*
	 * Send a device power IRP to our devnode.  Using our device object will
	 * only work on win2k and other NT based systems.
	 */
	status = PoRequestPowerIrp( p_dev_obj, IRP_MN_SET_POWER, state,
		__RequestPowerCompletion, p_irp, NULL );

	HCA_PRINT( TRACE_LEVEL_INFORMATION, HCA_DBG_PO, 
		("PoRequestPowerIrp: SET_POWER 'PowerDeviceD%d', status %#x\n", 
		state.DeviceState - 1, status ));

	if( status != STATUS_PENDING ) {
		HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PO,
			("PoRequestPowerIrp returned %08x.\n", status));
		p_irp->IoStatus.Status = status; 	/* Propagate the failure. */
		PoStartNextPowerIrp( p_irp );
		IoCompleteRequest( p_irp, IO_NO_INCREMENT );
		goto release;
	}

	status = STATUS_MORE_PROCESSING_REQUIRED;
	goto exit;

release:	
	IoReleaseRemoveLock( &p_fdo->cl_ext.remove_lock, p_irp );
exit:
	HCA_EXIT( HCA_DBG_PO );
	return status;
}


/* Work item callback to handle DevicePowerD0 IRPs at passive level. */
static void
__DevicePowerUpCompletionWorkItem(
	IN				DEVICE_OBJECT*				p_dev_obj,
	IN				void*						context )
{
	int					err;
	NTSTATUS			status;
	IO_STACK_LOCATION	*pIoStack;
	PFDO_DEVICE_DATA	p_fdo;
	IRP					*p_irp;
	POWER_STATE powerState;

	HCA_ENTER( HCA_DBG_PO );

	p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;
	p_irp = (IRP*)context;
	pIoStack = IoGetCurrentIrpStackLocation( p_irp );

	IoFreeWorkItem( p_fdo->pPoWorkItem );
	p_fdo->pPoWorkItem = NULL;

	/* restart the HCA */
	HCA_PRINT( TRACE_LEVEL_INFORMATION, HCA_DBG_PO, 
		("***** Restart the HCA, IRQL %d\n", KeGetCurrentIrql()));

	/* get MLX4_BUS IB interface */
	status = __get_ifc( p_dev_obj, &MLX4_BUS_IB_INTERFACE_GUID,
		sizeof(MLX4_BUS_IB_INTERFACE), MLX4_BUS_IB_INTERFACE_VERSION, NULL, (PINTERFACE)&p_fdo->bus_ib_ifc);
	if( !NT_SUCCESS( status ) ) {
		HCA_PRINT(TRACE_LEVEL_ERROR, HCA_DBG_PNP, ("Getting MLX4 BUS interface failed: status=0x%x\n", status));
		goto err_hca_reg;
	}
	p_fdo->bus_ib_ifc_taken = TRUE;
	p_fdo->bus_ib_ifc.p_ibdev->x.p_fdo = p_fdo;

	/* get node GUID */
	err = __get_dev_info( p_fdo, &p_fdo->hca.guid, &p_fdo->hca.hw_ver );
	if (err) {

		HCA_PRINT(TRACE_LEVEL_ERROR   ,HCA_DBG_LOW   ,
			("can't get guid - ib_query_device() failed (%08X)\n", err ));
		//TODO: no cleanup on error
		goto err_hca_reg;
	}

	p_fdo->DevicePowerState = pIoStack->Parameters.Power.State.DeviceState;
	powerState = PoSetPowerState( p_dev_obj, DevicePowerState,
		pIoStack->Parameters.Power.State );

	HCA_PRINT( TRACE_LEVEL_INFORMATION, HCA_DBG_PO, 
		("PoSetPowerState: old state %d, new state to %d\n", 
		powerState.DeviceState, p_fdo->DevicePowerState ));

	goto exit;

err_hca_reg:
	/* Flag device as having failed. */
	p_fdo->pnpState |= PNP_DEVICE_FAILED;
	IoInvalidateDeviceState( p_fdo->cl_ext.p_pdo );
exit:
	PoStartNextPowerIrp( p_irp );
	IoCompleteRequest( p_irp, IO_NO_INCREMENT );
	IoReleaseRemoveLock( &p_fdo->cl_ext.remove_lock, p_irp );
	HCA_EXIT( HCA_DBG_PO );
}

/*NOTE: Completion routines must NEVER be pageable. */
static NTSTATUS
__DevicePowerUpCompletion(
	IN				DEVICE_OBJECT				*p_dev_obj,
	IN				IRP							*p_irp,
	IN				void						*context )
{
	NTSTATUS			status = STATUS_SUCCESS;
	PFDO_DEVICE_DATA	p_fdo;
	IO_STACK_LOCATION	*pIoStack;

	HCA_ENTER( HCA_DBG_PO );

	UNUSED_PARAM( context );

	p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;
	pIoStack = IoGetCurrentIrpStackLocation( p_irp );

	if( !NT_SUCCESS( p_irp->IoStatus.Status ) ) {
		HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PO, 
			("IRP_MN_SET_POWER for device failed by lower driver with %08x.\n",
			p_irp->IoStatus.Status));
		status =  STATUS_SUCCESS;
		PoStartNextPowerIrp( p_irp );
		goto release;
	}

	/* Process in a work item - mthca_start blocks. */
	ASSERT( !p_fdo->pPoWorkItem );
	p_fdo->pPoWorkItem = IoAllocateWorkItem( p_dev_obj );
	if( !p_fdo->pPoWorkItem ) {
		HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PO, 
			("Failed to allocate work item.\n" ));
		status = STATUS_SUCCESS;
		p_fdo->pnpState |= PNP_DEVICE_FAILED;
		IoInvalidateDeviceState( p_fdo->cl_ext.p_pdo );
		PoStartNextPowerIrp( p_irp );
		goto release;
	}

	/* Process in work item callback. */
	IoMarkIrpPending( p_irp );
	IoQueueWorkItem( p_fdo->pPoWorkItem, 
		__DevicePowerUpCompletionWorkItem, DelayedWorkQueue, p_irp );
	status = STATUS_MORE_PROCESSING_REQUIRED;
	goto exit;

release:	
	IoReleaseRemoveLock( &p_fdo->cl_ext.remove_lock, p_irp );
exit:	
	HCA_EXIT( HCA_DBG_PO );
	return status;
}

static NTSTATUS __DevicePowerDownWorkItemCompletion(
	IN				DEVICE_OBJECT	*p_dev_obj,
	IN				IRP				*p_irp,
	IN				void				*context )
{
	PFDO_DEVICE_DATA  p_fdo  = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;
	UNUSED_PARAM( context );

	HCA_ENTER( HCA_DBG_PO );

	PoStartNextPowerIrp( p_irp );
	IoReleaseRemoveLock( &p_fdo->cl_ext.remove_lock, p_irp );

	HCA_EXIT( HCA_DBG_PO );
	return STATUS_SUCCESS;
}

/* Work item callback to handle DevicePowerD3 IRPs at passive level. */
static void
__DevicePowerDownWorkItem(
	IN				DEVICE_OBJECT*				p_dev_obj,
	IN				void*						context )
{
	IO_STACK_LOCATION	*pIoStack;
	PFDO_DEVICE_DATA	p_fdo;
	IRP					*p_irp;
	POWER_STATE powerState;

	HCA_ENTER( HCA_DBG_PO );

	p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;
	p_irp = (IRP*)context;
	pIoStack = IoGetCurrentIrpStackLocation( p_irp );

	IoFreeWorkItem( p_fdo->pPoWorkItem );
	p_fdo->pPoWorkItem = NULL;

	p_fdo->DevicePowerState = pIoStack->Parameters.Power.State.DeviceState;
	powerState = PoSetPowerState( p_dev_obj, DevicePowerState,
		pIoStack->Parameters.Power.State );

	HCA_PRINT( TRACE_LEVEL_INFORMATION, HCA_DBG_PO, 
		("PoSetPowerState: old state %d, new state to %d, IRQL %d\n", 
		powerState.DeviceState, p_fdo->DevicePowerState, KeGetCurrentIrql() ));

	HCA_PRINT( TRACE_LEVEL_INFORMATION, HCA_DBG_PO, 
		("***** Remove the HCA \n"));

	// release MLX4_BUS resources
	if(p_fdo->bus_ib_ifc_taken) {
		p_fdo->bus_ib_ifc_taken = FALSE;
		__put_ifc( (PINTERFACE)&p_fdo->bus_ib_ifc );
	}

	IoCopyCurrentIrpStackLocationToNext( p_irp );
#pragma warning( push, 3 )
	IoSetCompletionRoutine( p_irp, __DevicePowerDownWorkItemCompletion,
		NULL, TRUE, TRUE, TRUE );
#pragma warning( pop )
	PoCallDriver( p_fdo->cl_ext.p_next_do, p_irp );

	HCA_EXIT( HCA_DBG_PO );
}


static NTSTATUS
hca_set_power(
	IN				DEVICE_OBJECT* const		p_dev_obj,
	IN				IRP* const					p_irp,
		OUT			cl_irp_action_t* const		p_action )
{
	NTSTATUS			status;
	IO_STACK_LOCATION	*pIoStack;
	PFDO_DEVICE_DATA	p_fdo;

	HCA_ENTER( HCA_DBG_PO );

	p_fdo = (PFDO_DEVICE_DATA)p_dev_obj->DeviceExtension;
	pIoStack = IoGetCurrentIrpStackLocation( p_irp );

	HCA_PRINT( TRACE_LEVEL_INFORMATION, HCA_DBG_PO, 
		("SET_POWER for FDO %p (ext %p): type %s, state %d, action %d, IRQL %d \n",
		p_dev_obj, p_fdo,
		(pIoStack->Parameters.Power.Type) ? "DevicePowerState" : "SystemPowerState",
		pIoStack->Parameters.Power.State.DeviceState, 
		pIoStack->Parameters.Power.ShutdownType, KeGetCurrentIrql() ));

	switch( pIoStack->Parameters.Power.Type )
	{
	case SystemPowerState:
		p_fdo->SystemPowerState = pIoStack->Parameters.Power.State.SystemState;
		
		/*
		 * Process on the way up the stack.  We cannot block since the 
		 * power dispatch function can be called at elevated IRQL if the
		 * device is in a paging/hibernation/crash dump path.
		 */
		IoMarkIrpPending( p_irp );
		IoCopyCurrentIrpStackLocationToNext( p_irp );
#pragma warning( push, 3 )
		IoSetCompletionRoutine( p_irp, __SystemPowerCompletion, NULL, 
			TRUE, TRUE, TRUE );
#pragma warning( pop )
		PoCallDriver( p_fdo->cl_ext.p_next_do, p_irp );

		*p_action = IrpDoNothing;
		status = STATUS_PENDING;
		break;

	case DevicePowerState:
		IoMarkIrpPending( p_irp );
		if( pIoStack->Parameters.Power.State.DeviceState == PowerDeviceD0 && 
			p_fdo->SystemPowerState == PowerSystemWorking)
		{ /* power up */
			/* If we're already powered up, just pass down. */
			if( p_fdo->DevicePowerState == PowerDeviceD0 )
			{
				status = STATUS_SUCCESS;
				*p_action = IrpIgnore;
				break;
			}

			/* Process in I/O completion callback. */
			IoCopyCurrentIrpStackLocationToNext( p_irp );
#pragma warning( push, 3 )
			IoSetCompletionRoutine( p_irp, __DevicePowerUpCompletion, NULL, 
				TRUE, TRUE, TRUE );
#pragma warning( pop )
			PoCallDriver( p_fdo->cl_ext.p_next_do, p_irp );
		}
		else
		{ /* power down */

			/* Process in a work item - deregister_ca and HcaDeinit block. */
			ASSERT( !p_fdo->pPoWorkItem );
			p_fdo->pPoWorkItem = IoAllocateWorkItem( p_dev_obj );
			if( !p_fdo->pPoWorkItem )
			{
				status = STATUS_INSUFFICIENT_RESOURCES;
				break;
			}

			/* Process in work item callback. */
			IoQueueWorkItem(
				p_fdo->pPoWorkItem, __DevicePowerDownWorkItem, DelayedWorkQueue, p_irp );
		}
		*p_action = IrpDoNothing;
		status = STATUS_PENDING;
		break;

	default:
		/* Pass down and let the PDO driver handle it. */
		*p_action = IrpIgnore;
		status = STATUS_SUCCESS;
		break;
	}

	if( !NT_SUCCESS( status ) )
		*p_action = IrpComplete;

	HCA_EXIT( HCA_DBG_PNP );
	return status;
}

void
hca_init_vfptr( void )
{
	vfptrHcaPnp.identity = "MLX4_HCA driver";
	vfptrHcaPnp.pfn_start = hca_start;
	vfptrHcaPnp.pfn_query_stop = hca_query_stop;
	vfptrHcaPnp.pfn_stop = hca_stop;
	vfptrHcaPnp.pfn_cancel_stop = hca_cancel_stop;
	vfptrHcaPnp.pfn_query_remove = hca_query_remove;
	vfptrHcaPnp.pfn_release_resources = hca_release_resources;
	vfptrHcaPnp.pfn_remove = cl_do_remove;
	vfptrHcaPnp.pfn_cancel_remove = hca_cancel_remove;
	vfptrHcaPnp.pfn_surprise_remove = hca_surprise_remove;
	vfptrHcaPnp.pfn_query_capabilities = hca_query_capabilities;
	vfptrHcaPnp.pfn_query_pnp_state = hca_query_pnp_state;
	vfptrHcaPnp.pfn_filter_res_req = cl_irp_skip;
	vfptrHcaPnp.pfn_dev_usage_notification = cl_do_sync_pnp;
	vfptrHcaPnp.pfn_query_bus_relations = cl_irp_ignore;
	vfptrHcaPnp.pfn_query_ejection_relations = cl_irp_ignore;
	vfptrHcaPnp.pfn_query_removal_relations = cl_irp_ignore;
	vfptrHcaPnp.pfn_query_target_relations = cl_irp_ignore;
	vfptrHcaPnp.pfn_unknown = cl_irp_ignore;
	vfptrHcaPnp.pfn_query_resources = cl_irp_ignore;
	vfptrHcaPnp.pfn_query_res_req = cl_irp_ignore;
	vfptrHcaPnp.pfn_query_bus_info = cl_irp_ignore;
	vfptrHcaPnp.pfn_query_interface = hca_query_interface;
	vfptrHcaPnp.pfn_read_config = cl_irp_ignore;
	vfptrHcaPnp.pfn_write_config = cl_irp_ignore;
	vfptrHcaPnp.pfn_eject = cl_irp_ignore;
	vfptrHcaPnp.pfn_set_lock = cl_irp_ignore;
	vfptrHcaPnp.pfn_query_power = hca_query_power;
	vfptrHcaPnp.pfn_set_power = hca_set_power;
	vfptrHcaPnp.pfn_power_sequence = cl_irp_ignore;
	vfptrHcaPnp.pfn_wait_wake = cl_irp_ignore;
}

static NTSTATUS
__read_registry(
	IN				UNICODE_STRING* const	p_registry_path )
{
	NTSTATUS					status;
	/* Remember the terminating entry in the table below. */
	RTL_QUERY_REGISTRY_TABLE	table[3];

	HCA_ENTER( HCA_DBG_PNP );

	RtlInitUnicodeString( &g_param_path, NULL );
	g_param_path.MaximumLength = p_registry_path->Length + 
		sizeof(L"\\Parameters");
	g_param_path.Buffer = cl_zalloc( g_param_path.MaximumLength );
	if( !g_param_path.Buffer )
	{
		HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PNP, 
			("Failed to allocate parameters path buffer.\n"));
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	RtlAppendUnicodeStringToString( &g_param_path, p_registry_path );
	RtlAppendUnicodeToString( &g_param_path, L"\\Parameters" );

	/*
	 * Clear the table.  This clears all the query callback pointers,
	 * and sets up the terminating table entry.
	 */
	cl_memclr( table, sizeof(table) );

	/* Setup the table entries. */
	table[0].Flags = RTL_QUERY_REGISTRY_DIRECT;
	table[0].Name = L"DebugLevel";
	table[0].EntryContext = &g.DebugPrintLevel;
	table[0].DefaultType = REG_DWORD;
	table[0].DefaultData = &g.DebugPrintLevel;
	table[0].DefaultLength = sizeof(ULONG);

	
	table[1].Flags = RTL_QUERY_REGISTRY_DIRECT;
	table[1].Name = L"DebugFlags";
	table[1].EntryContext = &g.DebugPrintFlags;
	table[1].DefaultType = REG_DWORD;
	table[1].DefaultData = &g.DebugPrintFlags;
	table[1].DefaultLength = sizeof(ULONG);


	/* Have at it! */
	status = RtlQueryRegistryValues( RTL_REGISTRY_ABSOLUTE, 
		g_param_path.Buffer, table, NULL, NULL );

	HCA_EXIT( HCA_DBG_PNP );
	return status;
}

static void
hca_drv_unload(
	IN				PDRIVER_OBJECT			p_driver_obj )
{
	HCA_ENTER( HCA_DBG_PNP );

	UNUSED_PARAM( p_driver_obj );

	cl_free( g_param_path.Buffer );
	
	HCA_EXIT( HCA_DBG_PNP );
#if defined(EVENT_TRACING)
	WPP_CLEANUP(p_driver_obj);
#endif

}

static NTSTATUS
hca_sysctl(
	IN				PDEVICE_OBJECT				p_dev_obj,
	IN				PIRP						p_irp )
{
	NTSTATUS		status;
	PFDO_DEVICE_DATA	p_fdo;

	HCA_ENTER( HCA_DBG_PNP );

	p_fdo = p_dev_obj->DeviceExtension;

	IoSkipCurrentIrpStackLocation( p_irp );
	status = IoCallDriver( p_fdo->cl_ext.p_next_do, p_irp );

	HCA_EXIT( HCA_DBG_PNP );
	return status;
}


NTSTATUS
DriverEntry(
	IN				PDRIVER_OBJECT			p_driver_obj,
	IN				PUNICODE_STRING			p_registry_path )
{
	NTSTATUS			status;
#if defined(EVENT_TRACING)
	WPP_INIT_TRACING(p_driver_obj ,p_registry_path);
#endif
	// global initializations
	g.DebugPrintLevel = TRACE_LEVEL_VERBOSE;
	g.DebugPrintFlags = 0xffff;

	HCA_ENTER( HCA_DBG_PNP );

	status = mlnx_hcas_init();
	if( status  != STATUS_SUCCESS ) {
		HCA_PRINT( TRACE_LEVEL_ERROR ,HCA_DBG_PNP ,
			("mlnx_hcas_init returned %#x.\n", status));
		goto end;
	}

	status = __read_registry( p_registry_path );
	if( !NT_SUCCESS( status ) )
	{
		HCA_PRINT( TRACE_LEVEL_ERROR, HCA_DBG_PNP, 
			("__read_registry_path returned 0x%X.\n", status));
		return status;
	}

	/*leo:  init function table */
	hca_init_vfptr();
	
	p_driver_obj->MajorFunction[IRP_MJ_PNP] = cl_pnp;
	p_driver_obj->MajorFunction[IRP_MJ_POWER] = cl_power;
	p_driver_obj->MajorFunction[IRP_MJ_SYSTEM_CONTROL] = hca_sysctl;
	p_driver_obj->DriverUnload = hca_drv_unload;
	p_driver_obj->DriverExtension->AddDevice = hca_add_device;

end:
	HCA_EXIT( HCA_DBG_PNP );
	return STATUS_SUCCESS;
}

#endif
